"""
This type stub file was generated by pyright.
"""

from ruamel.yaml.error import MarkedYAMLError
from ruamel.yaml.tokens import *
from ruamel.yaml.compat import _debug

if False: ...
__all__ = ["Scanner", "RoundTripScanner", "ScannerError"]
_THE_END = ...
_THE_END_SPACE_TAB = ...
_SPACE_TAB = ...
if _debug != 0:
    def xprintf(*args: Any, **kw: Any) -> Any: ...

class ScannerError(MarkedYAMLError): ...

class SimpleKey:
    def __init__(
        self,
        token_number: Any,
        required: Any,
        index: int,
        line: int,
        column: int,
        mark: Any,
    ) -> None: ...

class Scanner:
    def __init__(self, loader: Any = ...) -> None:
        """Initialize the scanner."""
        ...

    @property
    def flow_level(self) -> int: ...
    def reset_scanner(self) -> None: ...
    @property
    def reader(self) -> Any: ...
    @property
    def scanner_processing_version(self) -> Any: ...
    def check_token(self, *choices: Any) -> bool: ...
    def peek_token(self) -> Any: ...
    def get_token(self) -> Any: ...
    def need_more_tokens(self) -> bool: ...
    def fetch_comment(self, comment: Any) -> None: ...
    def fetch_more_tokens(self) -> Any: ...
    def next_possible_simple_key(self) -> Any: ...
    def stale_possible_simple_keys(self) -> None: ...
    def save_possible_simple_key(self) -> None: ...
    def remove_possible_simple_key(self) -> None: ...
    def unwind_indent(self, column: Any) -> None: ...
    def add_indent(self, column: int) -> bool: ...
    def fetch_stream_start(self) -> None: ...
    def fetch_stream_end(self) -> None: ...
    def fetch_directive(self) -> None: ...
    def fetch_document_start(self) -> None: ...
    def fetch_document_end(self) -> None: ...
    def fetch_document_indicator(self, TokenClass: Any) -> None: ...
    def fetch_flow_sequence_start(self) -> None: ...
    def fetch_flow_mapping_start(self) -> None: ...
    def fetch_flow_collection_start(self, TokenClass: Any, to_push: Text) -> None: ...
    def fetch_flow_sequence_end(self) -> None: ...
    def fetch_flow_mapping_end(self) -> None: ...
    def fetch_flow_collection_end(self, TokenClass: Any) -> None: ...
    def fetch_flow_entry(self) -> None: ...
    def fetch_block_entry(self) -> None: ...
    def fetch_key(self) -> None: ...
    def fetch_value(self) -> None: ...
    def fetch_alias(self) -> None: ...
    def fetch_anchor(self) -> None: ...
    def fetch_tag(self) -> None: ...
    def fetch_literal(self) -> None: ...
    def fetch_folded(self) -> None: ...
    def fetch_block_scalar(self, style: Any) -> None: ...
    def fetch_single(self) -> None: ...
    def fetch_double(self) -> None: ...
    def fetch_flow_scalar(self, style: Any) -> None: ...
    def fetch_plain(self) -> None: ...
    def check_directive(self) -> Any: ...
    def check_document_start(self) -> Any: ...
    def check_document_end(self) -> Any: ...
    def check_block_entry(self) -> Any: ...
    def check_key(self) -> Any: ...
    def check_value(self) -> Any: ...
    def check_plain(self) -> Any: ...
    def scan_to_next_token(self) -> Any: ...
    def scan_directive(self) -> Any: ...
    def scan_directive_name(self, start_mark: Any) -> Any: ...
    def scan_yaml_directive_value(self, start_mark: Any) -> Any: ...
    def scan_yaml_directive_number(self, start_mark: Any) -> Any: ...
    def scan_tag_directive_value(self, start_mark: Any) -> Any: ...
    def scan_tag_directive_handle(self, start_mark: Any) -> Any: ...
    def scan_tag_directive_prefix(self, start_mark: Any) -> Any: ...
    def scan_directive_ignored_line(self, start_mark: Any) -> None: ...
    def scan_anchor(self, TokenClass: Any) -> Any: ...
    def scan_tag(self) -> Any: ...
    def scan_block_scalar(self, style: Any, rt: Optional[bool] = ...) -> Any: ...
    def scan_block_scalar_indicators(self, start_mark: Any) -> Any: ...
    def scan_block_scalar_ignored_line(self, start_mark: Any) -> Any: ...
    def scan_block_scalar_indentation(self) -> Any: ...
    def scan_block_scalar_breaks(self, indent: int) -> Any: ...
    def scan_flow_scalar(self, style: Any) -> Any: ...

    ESCAPE_REPLACEMENTS = ...
    ESCAPE_CODES = ...
    def scan_flow_scalar_non_spaces(self, double: Any, start_mark: Any) -> Any: ...
    def scan_flow_scalar_spaces(self, double: Any, start_mark: Any) -> Any: ...
    def scan_flow_scalar_breaks(self, double: Any, start_mark: Any) -> Any: ...
    def scan_plain(self) -> Any: ...
    def scan_plain_spaces(self, indent: Any, start_mark: Any) -> Any: ...
    def scan_tag_handle(self, name: Any, start_mark: Any) -> Any: ...
    def scan_tag_uri(self, name: Any, start_mark: Any) -> Any: ...
    def scan_uri_escapes(self, name: Any, start_mark: Any) -> Any: ...
    def scan_line_break(self) -> Any: ...

class RoundTripScanner(Scanner):
    def check_token(self, *choices: Any) -> bool: ...
    def peek_token(self) -> Any: ...
    def get_token(self) -> Any: ...
    def fetch_comment(self, comment: Any) -> None: ...
    def scan_to_next_token(self) -> Any: ...
    def scan_line_break(self, empty_line: bool = ...) -> Text: ...
    def scan_block_scalar(self, style: Any, rt: Optional[bool] = ...) -> Any: ...
    def scan_uri_escapes(self, name: Any, start_mark: Any) -> Any:
        """
        The roundtripscanner doesn't do URI escaping
        """
        ...

VALUECMNT = ...
KEYCMNT = ...

class CommentBase:
    __slots__ = ...
    def __init__(self, value: Any, line: Any, column: Any) -> None: ...
    def set_used(self, v: Any = ...) -> None: ...
    def set_assigned(self) -> None: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...
    def info(self) -> str: ...

class EOLComment(CommentBase):
    name = ...
    def __init__(self, value: Any, line: Any, column: Any) -> None: ...

class FullLineComment(CommentBase):
    name = ...
    def __init__(self, value: Any, line: Any, column: Any) -> None: ...

class BlankLineComment(CommentBase):
    name = ...
    def __init__(self, value: Any, line: Any, column: Any) -> None: ...

class ScannedComments:
    def __init__(self: Any) -> None: ...
    def add_eol_comment(self, comment: Any, column: Any, line: Any) -> Any: ...
    def add_blank_line(self, comment: Any, column: Any, line: Any) -> Any: ...
    def add_full_line_comment(self, comment: Any, column: Any, line: Any) -> Any: ...
    def __getitem__(self, idx: Any) -> Any: ...
    def __str__(self) -> Any: ...
    def last(self) -> str: ...
    def any_unprocessed(self) -> bool: ...
    def unprocessed(self, use: Any = ...) -> Any: ...
    def assign_pre(self, token: Any) -> Any: ...
    def assign_eol(self, tokens: Any) -> Any: ...
    def assign_post(self, token: Any) -> Any: ...
    def str_unprocessed(self) -> Any: ...

class RoundTripScannerSC(Scanner):
    def __init__(self, *arg: Any, **kw: Any) -> None: ...
    def get_token(self) -> Any: ...
    def need_more_tokens(self) -> bool: ...
    def scan_to_next_token(self) -> None: ...
    def scan_empty_or_full_line_comments(self) -> None: ...
    def scan_block_scalar_ignored_line(self, start_mark: Any) -> Any: ...



# type: ignore

try:
    from snakemake.cli import get_argument_parser, main
    from snakemake.common import configfile
    from snakemake.common.configfile import load_configfile
except ImportError:
    import snakemake.io as configfile
    from snakemake import get_argument_parser, main
    from snakemake.io import load_configfile

from snakemake.exceptions import WildcardError
from snakemake.io import expand
from snakemake.script import Snakemake

__all__ = [
    "Snakemake",
    "WildcardError",
    "configfile",
    "expand",
    "get_argument_parser",
    "load_configfile",
    "main",
]



# <AUTOGEN_INIT>
from ._config import (
    set_bids_spec,
)
from ._factory import (
    BidsFunction,
    bids_factory,
)
from ._presets import (
    bids,
)
from ._utils import (
    BidsPathEntitySpec,
    BidsPathSpec,
    BidsPathSpecFile,
)

__all__ = [
    "BidsFunction",
    "BidsPathEntitySpec",
    "BidsPathSpec",
    "BidsPathSpecFile",
    "bids",
    "bids_factory",
    "set_bids_spec",
]
# </AUTOGEN_INIT>



from __future__ import annotations

from collections.abc import Iterable


class ConfigError(Exception):
    """Exception raised for errors with the Snakebids config."""

    def __init__(self, msg: str) -> None:
        self.msg = msg
        super().__init__(msg)


class RunError(Exception):
    """Exception raised for errors in generating and running the snakemake workflow."""

    def __init__(self, msg: str, *args: object) -> None:
        super().__init__(msg, *args)
        self.msg = msg


class PybidsError(Exception):
    """Exception raised when pybids encounters a problem."""


class DuplicateComponentError(Exception):
    """Raised when a dataset is constructed from components with the same name."""

    def __init__(self, duplicated_names: Iterable[str]):
        self.duplicated_names_str = ", ".join(map(repr, duplicated_names))
        super().__init__(
            "A BidsDataset cannot be instantiated from BidsComponents with the same "
            "names. The following duplicate names were found: "
            f"{self.duplicated_names_str}."
        )


class SnakebidsPluginError(Exception):
    """Exception raised when a Snakebids plugin encounters a problem."""



"""
This type stub file was generated by pyright.
"""

from ruamel.yaml.error import YAMLError

if False: ...
__all__ = ["Serializer", "SerializerError"]

class SerializerError(YAMLError): ...

class Serializer:
    ANCHOR_TEMPLATE = ...
    ANCHOR_RE = ...
    def __init__(
        self,
        encoding: Any = ...,
        explicit_start: Optional[bool] = ...,
        explicit_end: Optional[bool] = ...,
        version: Optional[VersionType] = ...,
        tags: Any = ...,
        dumper: Any = ...,
    ) -> None: ...
    @property
    def emitter(self) -> Any: ...
    @property
    def resolver(self) -> Any: ...
    def open(self) -> None: ...
    def close(self) -> None: ...
    def serialize(self, node: Any) -> None: ...
    def anchor_node(self, node: Any) -> None: ...
    def generate_anchor(self, node: Any) -> Any: ...
    def serialize_node(self, node: Any, parent: Any, index: Any) -> None: ...

def templated_id(s: Text) -> Any: ...



from .bidsargs import (
    BidsArgs,
)
from .cli_config import (
    CliConfig,
)
from .component_edit import (
    ComponentEdit,
    FilterParse,
    FilterParseError,
)
from .pybidsdb import (
    Pybidsdb,
    logger,
)
from .snakemake import (
    SnakemakeBidsApp,
)
from .validator import (
    BidsValidator,
    InvalidBidsError,
)
from .version import (
    Version,
)

__all__ = [
    "BidsArgs",
    "BidsValidator",
    "CliConfig",
    "ComponentEdit",
    "FilterParse",
    "FilterParseError",
    "InvalidBidsError",
    "Pybidsdb",
    "SnakemakeBidsApp",
    "Version",
    "logger",
]



"""
This type stub file was generated by pyright.
"""

if False: ...
__all__ = ["ScalarInt", "BinaryInt", "OctalInt", "HexInt", "HexCapsInt", "DecimalInt"]

class ScalarInt(int):
    def __new__(cls: Any, *args: Any, **kw: Any) -> Any: ...
    def __iadd__(self, a: Any) -> Any: ...
    def __ifloordiv__(self, a: Any) -> Any: ...
    def __imul__(self, a: Any) -> Any: ...
    def __ipow__(self, a: Any) -> Any: ...
    def __isub__(self, a: Any) -> Any: ...
    @property
    def anchor(self) -> Any: ...
    def yaml_anchor(self, any: bool = ...) -> Any: ...
    def yaml_set_anchor(self, value: Any, always_dump: bool = ...) -> None: ...

class BinaryInt(ScalarInt):
    def __new__(
        cls, value: Any, width: Any = ..., underscore: Any = ..., anchor: Any = ...
    ) -> Any: ...

class OctalInt(ScalarInt):
    def __new__(
        cls, value: Any, width: Any = ..., underscore: Any = ..., anchor: Any = ...
    ) -> Any: ...

class HexInt(ScalarInt):
    """uses lower case (a-f)"""
    def __new__(
        cls, value: Any, width: Any = ..., underscore: Any = ..., anchor: Any = ...
    ) -> Any: ...

class HexCapsInt(ScalarInt):
    """uses upper case (A-F)"""
    def __new__(
        cls, value: Any, width: Any = ..., underscore: Any = ..., anchor: Any = ...
    ) -> Any: ...

class DecimalInt(ScalarInt):
    """needed if anchor"""
    def __new__(
        cls, value: Any, width: Any = ..., underscore: Any = ..., anchor: Any = ...
    ) -> Any: ...



"""File globbing functions based on snakemake.io library."""

from __future__ import annotations

import collections
import os
import re
from collections.abc import Sequence
from itertools import chain
from pathlib import Path

from snakebids.types import ZipList
from snakebids.utils.containers import MultiSelectDict


def regex(filepattern: str) -> str:
    """Build Snakebids regex based on the given file pattern."""
    regex_list: list[str] = []
    last = 0
    wildcards: set[str] = set()
    for match in _wildcard_regex.finditer(filepattern):
        regex_list.append(re.escape(filepattern[last : match.start()]))
        wildcard = match.group("name")
        if wildcard in wildcards:
            if match.group("constraint"):
                msg = (
                    "Constraint regex must be defined only in the first "
                    "occurrence of the wildcard in a string."
                )
                raise ValueError(msg)
            regex_list.append(f"(?P={wildcard})")
        else:
            wildcards.add(wildcard)
            constraint = (
                match.group("constraint") if match.group("constraint") else ".+"
            )
            regex_list.append(f"(?P<{wildcard}>{constraint})")

        last = match.end()
    regex_list.append(re.escape(filepattern[last:]))
    regex_list.append("$")  # ensure that the match spans the whole file
    return "".join(regex_list)


_wildcard_regex = re.compile(
    r"""
    \{
        (?=(   # This lookahead assertion emulates an 'atomic group'
               # which is required for performance
            \s*(?P<name>\w+)                    # wildcard name
            (\s*,\s*
                (?P<constraint>                 # an optional constraint
                    ([^{}]+ | \{\d+(,\d+)?\})*  # allow curly braces to nest
                )                               # ...  as in '{w,a{3,5}}'
            )?\s*
        ))\1
    \}
    """,
    re.VERBOSE,
)


def glob_wildcards(
    pattern: str | Path,
    files: Sequence[str | Path] | None = None,
    followlinks: bool = False,
) -> ZipList:
    """Glob the values of wildcards by matching a pattern to the filesystem.

    Returns a zip_list of field names with matched wildcard values.

    Parameters
    ----------
    pattern
        Path including wildcards to glob on the filesystem.
    files
        Files from which to glob wildcards. If None (default), the directory
        corresponding to the first wildcard in the pattern is walked, and
        wildcards are globbed from all files.
    followlinks
        Whether to follow links when globbing wildcards.
    """
    pattern = os.path.normpath(pattern)
    first_wildcard = re.search("{[^{]", pattern)
    dirname = (
        os.path.dirname(pattern[: first_wildcard.start()])
        if first_wildcard
        else os.path.dirname(pattern)
    )
    if not dirname:
        dirname = Path(".")

    names = [match.group("name") for match in _wildcard_regex.finditer(pattern)]

    # remove duplicates while preserving ordering
    names = list(dict.fromkeys(names))

    wildcards: dict[str, list[str]] = collections.defaultdict(list)

    re_pattern = re.compile(regex(pattern))

    file_iter = (
        (
            Path(dirpath, f)
            for dirpath, dirnames, filenames in os.walk(
                dirname, followlinks=followlinks
            )
            for f in chain(filenames, dirnames)
        )
        if files is None
        else iter(files)
    )

    for f in file_iter:
        if match := re.match(re_pattern, str(f)):
            for name, value in match.groupdict().items():
                wildcards[name].append(value)
    return MultiSelectDict(wildcards)


def update_wildcard_constraints(
    pattern: str,
    wildcard_constraints: dict[str, str],
    global_wildcard_constraints: dict[str, str],
) -> str:
    """Update wildcard constraints.

    Parameters
    ----------
    pattern : str
        Pattern on which to update constraints.
    wildcard_constraints : dict
        Dictionary of wildcard:constraint key-value pairs.
    global_wildcard_constraints : dict
        Dictionary of wildcard:constraint key-value pairs.
    """

    def replace_constraint(match: re.Match[str]):
        name = match.group("name")
        constraint = match.group("constraint")
        newconstraint = wildcard_constraints.get(
            name, global_wildcard_constraints.get(name)
        )
        if name in examined_names:
            return match.group(0)
        examined_names.add(name)
        # Don't override if constraint already set
        if constraint is not None:
            return match.group(0)
        # Only update if a new constraint has actually been set
        if newconstraint is not None:
            return f"{{{name},{newconstraint}}}"
        return match.group(0)

    examined_names: set[str] = set()
    return _wildcard_regex.sub(replace_constraint, pattern)



from __future__ import annotations

import io
from collections import OrderedDict
from io import StringIO
from pathlib import Path
from typing import Any

import pytest
from hypothesis import given
from hypothesis import strategies as st
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture
from pytest_mock.plugin import MockType

import snakebids.io.config as configio
import snakebids.io.yaml as yamlio
from tests.helpers import allow_function_scoped

YAML_SAFE_CHARS = st.characters(blacklist_characters=["\\"], blacklist_categories=["C"])


@given(path=st.text(YAML_SAFE_CHARS, min_size=1).map(Path))
def test_paths_formatted_as_str(path: Path):
    string = StringIO()
    yaml = yamlio.get_yaml_io()
    yaml.dump({"key": path}, string)
    string.seek(0, 0)
    assert yaml.load(string)["key"] == str(path)


class TestWriteConfig:
    def io_mocks(self, mocker: MockerFixture) -> dict[str, MockType]:
        return {
            "mopen": mocker.patch.object(configio, "open", mocker.mock_open()),  # pyright: ignore[reportUnknownMemberType]
            "jsondump": mocker.patch.object(configio.json, "dump"),
            "mkdir": mocker.patch.object(configio.Path, "mkdir"),
            "yamldump": mocker.patch.object(yamlio.YAML, "dump"),
        }

    @allow_function_scoped
    @given(
        ext=st.sampled_from([".json", ".yaml", ".yml"]),
        path=st.text().map(Path).filter(lambda p: p != Path() and not p.exists()),
    )
    def test_writes_correct_format(self, ext: str, path: Path, mocker: MockerFixture):
        mocker.stopall()
        mocks = self.io_mocks(mocker)
        path = path.with_suffix(ext)
        configio.write_config(path, {})
        if ext == ".json":
            mocks["mopen"].assert_called_once_with(path, "w", encoding="utf-8")
            mocks["jsondump"].assert_called_once()
            mocks["yamldump"].assert_not_called()
        else:
            mocks["mopen"].assert_not_called()
            mocks["jsondump"].assert_not_called()
            mocks["yamldump"].assert_called_once_with({}, path)

    @allow_function_scoped
    @given(
        path=st.text().filter(lambda s: Path(s).resolve() not in {Path(), Path("/")})
    )
    def test_doesnt_overwrite_file(self, path: str, fakefs: FakeFilesystem):
        fakefs.reset()
        fakefs.create_file(path)
        with pytest.raises(FileExistsError, match="already exists"):
            configio.write_config(path, {})

    @allow_function_scoped
    @given(
        path=st.text().filter(lambda s: Path(s).resolve() not in {Path(), Path("/")})
    )
    def test_overwrites_file_if_forced(
        self, path: str, fakefs: FakeFilesystem, mocker: MockerFixture
    ):
        mocker.stopall()
        fakefs.reset()
        fakefs.create_file(path)
        mocks = self.io_mocks(mocker)
        configio.write_config(path, {}, force_overwrite=True)
        assert mocks["jsondump"].call_count ^ mocks["yamldump"].call_count

    @given(
        data=st.recursive(
            st.text(YAML_SAFE_CHARS),
            lambda children: st.dictionaries(st.text(YAML_SAFE_CHARS), children),
        )
    )
    def test_ordered_dict_roundtrips_as_dict(self, data: dict[str, Any]):
        def to_odict(data: dict[str, Any] | str) -> OrderedDict[str, Any] | str:
            if isinstance(data, str):
                return data
            return OrderedDict({key: to_odict(value) for key, value in data.items()})

        stream = io.StringIO()
        yaml = yamlio.get_yaml_io()
        odict = to_odict(data)
        yaml.dump(odict, stream)
        stream.seek(0)
        assert yaml.load(stream) == data



"""Tests for snakemake_io"""

from snakebids.utils import snakemake_io


def test_glob_wildcards():
    """Test glob_wildcards() with various patterns"""
    file_path = "tests/data/bids_t1w/sub-001/anat/sub-001_acq-mprage_T1w.nii.gz"
    empty_wildcards = {}
    assert snakemake_io.glob_wildcards(file_path) == empty_wildcards

    acq_wildcard_path = "tests/data/bids_t1w/sub-001/anat/sub-001_acq-{acq}_T1w.nii.gz"
    acq_wildcards = {"acq": ["mprage"]}
    assert snakemake_io.glob_wildcards(acq_wildcard_path) == acq_wildcards

    # Order of wildcards in the namedtuple is not deterministic
    both_wildcard_path = (
        "tests/data/bids_t1w/sub-{subject}/anat/sub-{subject}_acq-{acq}_T1w.nii.gz"
    )
    both_wildcards = [
        {"subject": ["001", "002"], "acq": ["mprage", "mprage"]},
        {"subject": ["002", "001"], "acq": ["mprage", "mprage"]},
        {"acq": ["mprage", "mprage"], "subject": ["001", "002"]},
        {"acq": ["mprage", "mprage"], "subject": ["002", "001"]},
    ]
    both_wildcards_one_file = [
        {"subject": ["001"], "acq": ["mprage"]},
        {"acq": ["mprage"], "subject": ["001"]},
    ]

    assert snakemake_io.glob_wildcards(both_wildcard_path) in both_wildcards
    assert (
        snakemake_io.glob_wildcards(both_wildcard_path, files=[file_path])
        in both_wildcards_one_file
    )



"""
This type stub file was generated by pyright.
"""

import datetime

if False: ...

class LazyEval:
    """
    Lightweight wrapper around lazily evaluated func(*args, **kwargs).

    func is only evaluated when any attribute of its return value is accessed.
    Every attribute access is passed through to the wrapped value.
    (This only excludes special cases like method-wrappers, e.g., __hash__.)
    The sole additional attribute is the lazy_self function which holds the
    return value (or, prior to evaluation, func and arguments), in its closure.
    """
    def __init__(self, func: Callable[..., Any], *args: Any, **kwargs: Any) -> None: ...
    def __getattribute__(self, name: str) -> Any: ...
    def __setattr__(self, name: str, value: Any) -> None: ...

RegExp = ...
timestamp_regexp = ...

def create_timestamp(
    year: Any,
    month: Any,
    day: Any,
    t: Any,
    hour: Any,
    minute: Any,
    second: Any,
    fraction: Any,
    tz: Any,
    tz_sign: Any,
    tz_hour: Any,
    tz_minute: Any,
) -> Union[datetime.datetime, datetime.date]: ...
def load_yaml_guess_indent(stream: StreamTextType, **kw: Any) -> Any:
    """guess the indent and block sequence indent of yaml stream/string

    returns round_trip_loaded stream, indent level, block sequence indent
    - block sequence indent is the number of spaces before a dash relative to previous indent
    - if there are no block sequences, indent is taken from nested mappings, block sequence
      indent is unset (None) in that case
    """
    ...

def configobj_walker(cfg: Any) -> Any:
    """
    walks over a ConfigObj (INI file with comments) generating
    corresponding YAML output (including comments
    """
    ...



#!/usr/bin/env python3
from pathlib import Path

from snakebids import bidsapp, plugins

app = bidsapp.app(
    [
        plugins.SnakemakeBidsApp(Path(__file__).resolve().parent),
    ]
)

if __name__ == "__main__":
    app.run()



from __future__ import annotations

import argparse
import logging
from pathlib import Path
from typing import Any

import attrs

from snakebids import bidsapp
from snakebids.bidsapp.args import ArgumentGroups
from snakebids.plugins.base import PluginBase
from snakebids.utils.utils import DEPRECATION_FLAG

logger = logging.getLogger(__name__)


@attrs.define
class Pybidsdb(PluginBase):
    """Add CLI parameters to specify and reset a pybids database.

    Parameters
    ----------
    argument_group
        Specify title of the group to which arguments should be added


    CLI Arguments
    ~~~~~~~~~~~~~
    Two arguments are added to the CLI. These can be overridden by adding arguments
    with corresponding ``dests`` before this plugin is run:

    - ``plugins.pybidsdb.dir``: (:class:`~pathlib.Path`) Path of the database
    - ``plugins.pybidsdb.reset``: (:class:`bool`) Boolean indicating the database should
      be reset.

    After parsing, the above dests will be moved into ``config`` under the following
    names:

    - ``plugins.pybidsdb.dir`` → ``pybidsdb_dir``
    - ``plugins.pybidsdb.reset`` → ``pybidsdb_reset``

    This plugin only handles the CLI arguments, it does not do any actions with the
    database. The above config entries can be consumed by downstream processes.
    """

    argument_group: str | None = None

    PREFIX = "plugins.pybidsdb"

    def __eq__(self, other: Any):
        return isinstance(other, self.__class__)

    @bidsapp.hookimpl
    def add_cli_arguments(
        self, parser: argparse.ArgumentParser, argument_groups: ArgumentGroups
    ):
        """Add database parameters."""
        group = (
            argument_groups[self.argument_group]
            if self.argument_group is not None
            else parser
        )
        self.try_add_argument(
            group,
            "--pybidsdb-dir",
            "--pybidsdb_dir",
            action="store",
            type=Path,
            dest="dir",
            metavar="PATH",
            help=(
                "Optional path to directory of SQLite databasefile for PyBIDS. "
                "If directory is passed and folder exists, indexing is skipped. "
                "If pybidsdb_reset is called, indexing will persist"
            ),
        )

        self.try_add_argument(
            group,
            "--pybidsdb-reset",
            "--pybidsdb_reset",
            action="store_true",
            dest="reset",
            help="Reindex existing PyBIDS SQLite database",
        )

        # To be deprecated
        self.try_add_argument(
            group,
            "--reset-db",
            "--reset_db",
            action="store_true",
            dest="reset_db",
            help=argparse.SUPPRESS,
        )

    @bidsapp.hookimpl
    def update_cli_namespace(self, namespace: dict[str, Any], config: dict[str, Any]):
        """Assign database parameters to config."""
        # Update config with pybids settings
        reset = self.pop(namespace, "reset")
        reset_db = self.pop(namespace, "reset_db")
        if reset_db:
            logger.warning(
                "--reset-db/--reset_db will be deprecated in a future release. To "
                "reset the pybids database, use the new --pybidsdb-reset flag instead."
            )
        config["pybidsdb_reset"] = reset or reset_db
        pybidsdb_dir = self.pop(namespace, "dir")

        config["pybidsdb_dir"] = (
            None if pybidsdb_dir is None else pybidsdb_dir.resolve()
        )
        config["pybids_db_dir"] = (
            f"{DEPRECATION_FLAG}{config['pybidsdb_dir']}{DEPRECATION_FLAG}"
            if config["pybidsdb_dir"] is not None
            else None
        )
        config["pybids_db_reset"] = (
            f"{DEPRECATION_FLAG}{int(config['pybidsdb_reset'])}{DEPRECATION_FLAG}"
        )



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, TypeAlias, TypedDict

if TYPE_CHECKING:
    from argparse import _ArgumentGroup

__all__ = ["ArgumentGroups", "SnakebidsConfig"]

ArgumentGroups: TypeAlias = "dict[str, _ArgumentGroup]"


class SnakebidsConfig(TypedDict):
    """Interface of default arguments parsed by {func}`snakebids.bidsapp.parser`."""

    bids_dir: Path
    output_dir: Path
    analysis_level: str
    participant_label: list[str] | None
    exclude_participant_label: list[str] | None
    derivatives: list[str] | None



from __future__ import annotations

from collections.abc import Iterable, Mapping, Sequence
from enum import Enum
from pathlib import Path
from typing import Generic, Protocol, TypeAlias, overload

from typing_extensions import Self, TypedDict, TypeVar

from snakebids.utils.containers import MultiSelectDict

_T_contra = TypeVar("_T_contra", contravariant=True)
_S_co = TypeVar("_S_co", covariant=True)


class FilterSpec(TypedDict, total=False):
    """Optional filter specification allowing regex matching."""

    get: str | bool | Sequence[str | bool]
    match: str
    search: str


FilterValue: TypeAlias = "str | bool | Sequence[str | bool] | FilterSpec"
FilterMap: TypeAlias = "Mapping[str, FilterValue]"


class InputConfig(TypedDict, total=False):
    """Configuration for a single bids component."""

    filters: FilterMap
    """Filters to pass on to :class:`BIDSLayout.get() <bids.layout.BIDSLayout>`

    Each key refers to the name of an entity. Values may take the following forms:

    * :class:`string <str>`: Restricts the entity to the exact string given
    * :class:`bool`: ``True`` requires the entity to be present (with any value).
      ``False`` requires the entity to be absent.
    * :class:`list` [:class:`str`]: List of allowable values the entity may take.

    In addition, a few special filters may be added which carry different meanings:

    * ``use_regex: True``: If present, all strings will be interpreted as regex
    * ``scope``: Restricts the scope of the component. It may take the following values:
        - ``"all"``: search everything (default behaviour)
        - ``"raw"``: only search the top-level raw dataset
        - ``"derivatives"``: only search derivative datasets
        - ``<PipelineName>``: only search derivative datasets with a matching pipeline
            name
    """

    wildcards: list[str]
    """Wildcards to allow in the component.

    Each value in the list refers to the name of an entity. If the entity is present,
    the generated :class:`~snakebids.BidsComponent` will have values of this entity
    substituted for wildcards in the :attr:`~snakebids.BidsComponent.path`, and the
    entity will be included in the :attr:`~snakebids.BidsComponent.zip_lists`.

    If the entity is not found, it will be ignored.
    """
    custom_path: str


class BinaryOperator(Protocol, Generic[_T_contra, _S_co]):
    """Callables that act on two objects of identical type."""

    def __call__(self, first: _T_contra, second: _T_contra, /) -> _S_co:  # noqa: D102
        ...


class Expandable(Protocol):
    """Protocol represents objects that hold an entity table and can expand over a path.

    Includes BidsComponent, BidsPartialComponent, and BidsComponentRow
    """

    @property
    def zip_lists(self) -> ZipList:  # noqa: D102
        ...

    def expand(  # noqa: D102
        self,
        paths: Iterable[Path | str] | Path | str,
        /,
        allow_missing: bool | str | Iterable[str] = False,
        **wildcards: str | Iterable[str],
    ) -> list[str]: ...

    def filter(  # noqa: D102
        self,
        *,
        regex_search: bool | str | Iterable[str] = False,
        **filters: str | Iterable[str],
    ) -> Self: ...


_K_contra = TypeVar("_K_contra", bound="str", contravariant=True)
_V_co = TypeVar("_V_co", covariant=True)
_Valt_co = TypeVar("_Valt_co", covariant=True)


class MultiSelectable(Protocol, Generic[_K_contra, _V_co, _Valt_co]):
    """Mappings supporting selection with multiple keys."""

    @overload
    def __getitem__(self, key: _K_contra, /) -> _V_co: ...

    @overload
    def __getitem__(self, key: tuple[_K_contra, ...], /) -> _Valt_co: ...


InputsConfig: TypeAlias = dict[str, InputConfig]
"""Configuration for all bids components to be parsed in the app

Should be defined in the config.yaml file, by convention in a key called 'pybids_inputs'
"""

ZipList: TypeAlias = MultiSelectDict[str, list[str]]
"""Multiselectable dict mapping entity names to possible values.

All lists must be the same length. Entries in each list with the same index correspond
to the same path. Thus, the ZipList can be read like a table, where each row corresponds
to an entity, and each "column" corresponds to a path.
"""


ZipListLike: TypeAlias = Mapping[str, Sequence[str]]
"""
Generic form of a :py:data:`ZipList`

Useful for typing functions that won't mutate the ZipList or use
:class:`~snakebids.utils.utils.MultiSelectDict` capabilities. Like :class:`ZipList`,
each :class:`~typing.Sequence` must be the same length, and values in each with the same
index must correspond to the same path.
"""


class OptionalFilterType(Enum):
    """Sentinel value for CLI OPTIONAL filtering.

    This is necessary because None means no CLI filter was added.
    """

    OptionalFilter = "OptionalFilter"


OptionalFilter = OptionalFilterType.OptionalFilter



"""
This type stub file was generated by pyright.
"""

from ruamel.yaml.error import *

if False: ...
__all__ = ["BaseResolver", "Resolver", "VersionedResolver"]
implicit_resolvers = ...

class ResolverError(YAMLError): ...

class BaseResolver:
    DEFAULT_SCALAR_TAG = ...
    DEFAULT_SEQUENCE_TAG = ...
    DEFAULT_MAPPING_TAG = ...
    yaml_implicit_resolvers: Dict[Any, Any] = ...
    yaml_path_resolvers: Dict[Any, Any] = ...
    def __init__(self: Any, loadumper: Any = ...) -> None: ...
    @property
    def parser(self) -> Any: ...
    @classmethod
    def add_implicit_resolver_base(cls, tag: Any, regexp: Any, first: Any) -> None: ...
    @classmethod
    def add_implicit_resolver(cls, tag: Any, regexp: Any, first: Any) -> None: ...
    @classmethod
    def add_path_resolver(cls, tag: Any, path: Any, kind: Any = ...) -> None: ...
    def descend_resolver(self, current_node: Any, current_index: Any) -> None: ...
    def ascend_resolver(self) -> None: ...
    def check_resolver_prefix(
        self, depth: int, path: Any, kind: Any, current_node: Any, current_index: Any
    ) -> bool: ...
    def resolve(self, kind: Any, value: Any, implicit: Any) -> Any: ...
    @property
    def processing_version(self) -> Any: ...

class Resolver(BaseResolver): ...

class VersionedResolver(BaseResolver):
    """
    contrary to the "normal" resolver, the smart resolver delays loading
    the pattern matching rules. That way it can decide to load 1.1 rules
    or the (default) 1.2 rules, that no longer support octal without 0o, sexagesimals
    and Yes/No/On/Off booleans.
    """
    def __init__(
        self,
        version: Optional[VersionType] = ...,
        loader: Any = ...,
        loadumper: Any = ...,
    ) -> None: ...
    def add_version_implicit_resolver(
        self, version: VersionType, tag: Any, regexp: Any, first: Any
    ) -> None: ...
    def get_loader_version(self, version: Optional[VersionType]) -> Any: ...
    @property
    def versioned_resolver(self) -> Any:
        """
        select the resolver based on the version we are parsing
        """
        ...

    def resolve(self, kind: Any, value: Any, implicit: Any) -> Any: ...
    @property
    def processing_version(self) -> Any: ...



from __future__ import annotations

import operator as op
import re
import string
import sys
from typing import Any

import more_itertools as itx
import pytest
from hypothesis import assume, given
from hypothesis import strategies as st

import tests.strategies as sb_st
from snakebids.utils.containers import ImmutableList, MultiSelectDict, RegexContainer
from snakebids.utils.utils import get_wildcard_dict, matches_any


@st.composite
def non_matching_match_list(draw: st.DrawFn) -> tuple[str, list[str]]:
    item = draw(st.text())
    match_list = draw(
        st.lists(
            st.text(
                # blank match strings (e.g. ['']) will match anything, and are not
                # considered here
                min_size=1
            )
            .map(re.escape)
            .filter(lambda s: not re.match(s, item))
        )
    )
    return item, match_list


class TestMatchesAny:
    @given(st.text(), st.lists(st.text()))
    def test_with_eq_operator(self, item: str, match_list: list[str]):
        if item not in match_list:
            assert matches_any(item, match_list, op.eq) is False
            match_list.append(item)
        assert matches_any(item, match_list, op.eq)

    @given(non_matching_match_list())
    def test_with_re_match(self, args: tuple[str, list[str]]):
        item, match_list = args

        if item not in match_list:
            assert matches_any(item, match_list, re.match) is False
            match_list.append(re.escape(item))
        assert matches_any(item, match_list, re.match)

    @given(st.emails())
    def test_with_email(self, email: str):
        match_list = ["non[^@]*?email[pattern]"]
        assert matches_any(email, match_list, re.match) is False
        # Email regex copied from https://www.emailregex.com/
        match_list.append(
            r'(?:[a-z0-9!#$%&\'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&\'*+/=?^_`{|}~-]+)*|"(?:'
            r"[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]|\\[\x01-\x09\x0b\x0c"
            r'\x0e-\x7f])*")@(?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-'
            r"9-]*[a-z0-9])?|\[(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0"
            r"-5]|2[0-4][0-9]|[01]?[0-9][0-9]?|[a-z0-9-]*[a-z0-9]:(?:[\x01-\x08\x0b\x0c"
            r"\x0e-\x1f\x21-\x5a\x53-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])+)\])"
        )
        assert matches_any(email, match_list, re.match, re.IGNORECASE)


class TestMultiselectDict:
    def get_selectors(
        self,
        data: st.DataObject,
        dicts: MultiSelectDict[str, Any],
        use_nonexistant_keys: bool = False,
        unique: bool = False,
    ) -> tuple[str, ...]:
        if not dicts and not use_nonexistant_keys:
            return ()
        sampler = st.sampled_from(list(dicts))
        val_strat = (
            st.text().filter(lambda s: s not in dicts)
            if use_nonexistant_keys
            else sampler
        )
        return tuple(data.draw(st.lists(val_strat, unique=unique)))

    @given(dicts=sb_st.multiselect_dicts(st.text(), sb_st.everything()), data=st.data())
    def test_multiple_selection_returns_same_type(
        self, dicts: MultiSelectDict[str, Any], data: st.DataObject
    ):
        selectors = self.get_selectors(data, dicts)
        # `noqa`, because in this case we actually want the classes to be exactly the
        # same
        assert type(dicts) == type(dicts[selectors])  # noqa: E721

    @given(dicts=sb_st.multiselect_dicts(st.text(), sb_st.everything()), data=st.data())
    def test_selected_items_in_original(
        self, dicts: MultiSelectDict[str, Any], data: st.DataObject
    ):
        selectors = self.get_selectors(data, dicts)
        selected = dicts[selectors]
        for key in selected:
            assert selected[key] is dicts[key]

    @given(dicts=sb_st.multiselect_dicts(st.text(), sb_st.everything()), data=st.data())
    def test_all_requested_items_received_and_no_others(
        self, dicts: MultiSelectDict[str, Any], data: st.DataObject
    ):
        selectors = self.get_selectors(data, dicts)
        selected = dicts[selectors]
        assert set(selectors) == set(selected)

    @given(
        dicts=sb_st.multiselect_dicts(st.text(), sb_st.everything(), min_size=1),
        data=st.data(),
    )
    def test_single_key_gives_single_item(
        self, dicts: MultiSelectDict[str, Any], data: st.DataObject
    ):
        selector = data.draw(st.sampled_from(list(dicts)))
        assert not isinstance(dicts[selector], MultiSelectDict)

    @given(
        dicts=sb_st.multiselect_dicts(st.text(), sb_st.everything(), min_size=1),
        data=st.data(),
    )
    def test_single_missing_key_raises_error(
        self, dicts: MultiSelectDict[str, Any], data: st.DataObject
    ):
        selector = data.draw(st.text().filter(lambda s: s not in dicts))
        with pytest.raises(KeyError):
            dicts[selector]

    @given(dicts=sb_st.multiselect_dicts(st.text(), sb_st.everything()), data=st.data())
    def test_multiple_missing_key_raises_error(
        self, dicts: MultiSelectDict[str, Any], data: st.DataObject
    ):
        selectors = self.get_selectors(data, dicts, use_nonexistant_keys=True)
        assume(len(selectors))
        with pytest.raises(KeyError):
            dicts[selectors]

    @given(dicts=sb_st.multiselect_dicts(st.text(), sb_st.everything()), data=st.data())
    def test_order_of_selectors_is_preserved(
        self, dicts: MultiSelectDict[str, Any], data: st.DataObject
    ):
        selectors = self.get_selectors(data, dicts)
        # Using the itx method for uniqueness to avoid calculating unique values in the
        # test in the same way as the source code
        assert tuple(dicts[selectors]) == tuple(itx.unique_everseen(selectors))


class TestImmutableListsAreEquivalentToTuples:
    @given(st.lists(sb_st.everything()))
    def test_they_are_not_tuples(self, items: list[Any]):
        iml = ImmutableList(items)
        assert not isinstance(iml, tuple)

    @given(st.lists(sb_st.partially_ordered(), min_size=1), st.data())
    def test_contains_items(self, items: list[Any], data: st.DataObject):
        iml = ImmutableList(items)
        sample = data.draw(st.sampled_from(items))
        assert sample in iml

    @given(st.lists(sb_st.hashables()))
    def test_is_hashable(self, items: list[Any]):
        iml = ImmutableList(items)
        assert hash(iml) == hash(tuple(items))

    @given(st.lists(sb_st.everything()))
    def test_is_iterable(self, items: list[Any]):
        i = 0
        iml = ImmutableList(items)
        _iter = iter(iml)
        while True:
            try:
                val = next(_iter)
                assert val is items[i]
                i += 1
            except StopIteration:
                break
        assert i == len(items)

    @given(st.lists(sb_st.everything()))
    def test_is_reversable(self, items: list[Any]):
        i = 0
        iml = ImmutableList(items)
        _iter = reversed(iml)
        while True:
            try:
                val = next(_iter)
                i += 1
                assert val is items[-i]
            except StopIteration:
                break
        assert i == len(items)

    @given(st.lists(sb_st.everything()))
    def test_has_length(self, items: list[Any]):
        iml = ImmutableList(items)
        assert len(iml) == len(items)

    @given(st.lists(sb_st.everything(), min_size=1), st.data())
    def test_supports_getting(self, items: list[Any], data: st.DataObject):
        iml = ImmutableList(items)
        index = data.draw(st.integers(min_value=0, max_value=len(items) - 1))
        assert iml[index] is items[index]

    @given(st.lists(sb_st.everything()), st.data())
    def test_supports_slicing(self, items: list[Any], data: st.DataObject):
        iml = ImmutableList(items)
        _slice = data.draw(st.slices(len(items)))
        for i, item in enumerate(iml[_slice]):
            assert item is items[_slice][i]

    @given(st.lists(st.integers()), st.lists(st.integers()))
    def test_supports_comparisons(self, items1: list[Any], items2: list[Any]):
        iml1 = ImmutableList(items1)
        iml2 = ImmutableList(items2)
        tup1 = tuple(items1)
        tup2 = tuple(items2)
        if tup1 == tup2:
            assert iml1 == iml2
            assert iml1 == tup2
            assert tup1 == iml2
        if tup1 >= tup2:
            assert iml1 >= iml2
            assert iml1 >= tup2
            assert tup1 >= iml2
        if tup1 > tup2:
            assert iml1 > iml2
            assert iml1 > tup2
            assert tup1 > iml2
        if tup1 < tup2:
            assert iml1 < iml2
            assert iml1 < tup2
            assert tup1 < iml2
        if tup1 <= tup2:
            assert iml1 <= iml2
            assert iml1 <= tup2
            assert tup1 <= iml2

    @given(st.lists(sb_st.partially_ordered()), st.lists(sb_st.partially_ordered()))
    def test_supports_equality(self, items1: list[Any], items2: list[Any]):
        iml1 = ImmutableList(items1)
        iml2 = ImmutableList(items2)
        tup1 = tuple(items1)
        tup2 = tuple(items2)
        if tup1 == tup2:
            assert iml1 == iml2
            assert iml1 == tup2
            assert tup1 == iml2

    @given(st.lists(sb_st.partially_ordered()))
    def test_equal_items_are_equal(self, items: list[Any]):
        iml1 = ImmutableList(items)
        iml2 = ImmutableList(items)
        tup1 = tuple(items)
        tup2 = tuple(items)
        assert iml1 == iml2
        assert iml1 == tup2
        assert tup1 == iml2

    @given(st.lists(sb_st.everything()))
    def test_supports_addition(self, items: list[Any]):
        iml = ImmutableList(items)
        assert iml + iml == tuple(items) + tuple(items)

    @given(
        st.lists(sb_st.everything(), max_size=10),
        st.integers(min_value=-5, max_value=5),
    )
    def test_supports_multiplication(self, items: list[Any], value: int):
        iml = ImmutableList(items)
        assert iml * value == tuple(items) * value
        assert value * iml == value * tuple(items)
        assert value * iml == iml * value

    @given(st.lists(sb_st.partially_ordered(), min_size=1), st.data())
    def test_supports_count(self, items: list[Any], data: st.DataObject):
        iml = ImmutableList(items)
        sample = data.draw(st.sampled_from(items))
        assert iml.count(sample) == items.count(sample)

    @given(st.lists(sb_st.partially_ordered(), min_size=1), st.data())
    def test_supports_index(self, items: list[Any], data: st.DataObject):
        iml = ImmutableList(items)
        _slice = data.draw(st.slices(len(items)))
        assume(len(items[_slice]))
        assume(_slice.step is None or _slice.step > 0)
        sample = data.draw(st.sampled_from(items[_slice]))
        start = 0 if _slice.start is None else _slice.start
        stop = sys.maxsize if _slice.stop is None else _slice.stop
        assert iml.index(sample, start, stop) == items.index(sample, start, stop)

    @given(st.lists(sb_st.everything()), sb_st.everything())
    def test_cannot_be_mutated(self, items: list[Any], value: Any):
        iml = ImmutableList(items)
        with pytest.raises(TypeError):
            iml[len(iml)] = value  # type: ignore

    @given(st.lists(st.text(string.printable)))
    def test_repr(self, items: list[Any]):
        iml = ImmutableList(items)
        assert repr(iml) == f"ImmutableList({items})"

    @given(st.lists(sb_st.everything()))
    def test_bool(self, items: list[Any]):
        iml = ImmutableList(items)
        assert bool(iml) == bool(items)


@given(
    st.dictionaries(
        sb_st.bids_entity().map(lambda e: e.wildcard),
        sb_st.bids_value("[^.]*"),
        min_size=1,
    ).filter(lambda v: list(v) != ["datatype"])
)
def test_get_wildcard_dict(zip_list: dict[str, str]):
    wildcards = get_wildcard_dict(zip_list)
    wildstr = ".".join(wildcards.values())
    first = wildstr.format(**wildcards)
    second = first.format(**zip_list)
    assert set(second.split(".")) == set(zip_list.values())


class TestRegexContainer:
    DDWW = r"^\d{3}[a-zA-Z]{3}$"
    bDDWW = rb"^\d{3}[a-zA-Z]{3}$"  # noqa: N815
    WWDD = r"^[a-zA-Z]{3}\d{3}$"

    @given(sample=st.from_regex(DDWW))
    def test_str_matching_regex_are_contained(self, sample: str):
        assert sample in RegexContainer(self.DDWW)

    @given(sample=st.from_regex(bDDWW))
    def test_bytes_matching_regex_are_contained(self, sample: bytes):
        assert sample in RegexContainer(self.bDDWW)

    @given(sample=st.from_regex(WWDD))
    def test_str_not_matching_regex_excluded(self, sample: str):
        assert sample not in RegexContainer(self.DDWW)

    @given(sample=st.from_regex(bDDWW))
    def test_container_specific_to_type(self, sample: str):
        assert sample not in RegexContainer(self.DDWW)



# ruff: noqa: PLR2004

import json
from pathlib import Path

import pytest

from snakebids.exceptions import RunError
from snakebids.utils import output


def dirlen(f: Path):
    return len([*f.iterdir()])


@pytest.fixture
def fake_snakemake(tmp_path: Path):
    app1 = tmp_path / "app1"
    app1.mkdir()
    (app1 / "random_file").touch()
    (app1 / ".snakebids").touch()
    (app1 / "config").mkdir()
    (app1 / "config" / "config.yaml").touch()

    (app1 / "results" / "app1").mkdir(parents=True)
    (app1 / "results" / "app1" / "result1.data").touch()
    (app1 / "results" / "app1" / "link").symlink_to(
        app1 / "workflow" / "rules" / "rule1.smk"
    )

    (app1 / "workflow" / "rules").mkdir(parents=True)
    (app1 / "workflow" / "Snakefile").touch()
    (app1 / "workflow" / "rules" / "rule1.smk").touch()

    (app1 / ".snakemake").mkdir()

    output = tmp_path / "output"
    output.mkdir()
    (output / "random_file").touch()
    (output / ".snakebids").touch()
    (output / "app1").mkdir(parents=True)
    (output / "app1" / "special_results.data").touch()

    old_style = tmp_path / "old-style"
    old_style.mkdir()
    (old_style / "config").mkdir()
    (old_style / "config" / "config.yaml").mkdir()
    (old_style / "app1").mkdir(parents=True)
    (old_style / "app1" / "special_results.data").touch()

    randomdir = tmp_path / "randomdir"
    randomdir.mkdir()
    (randomdir / "somefile").touch()

    return tmp_path


class TestGetSnakebidsFile:
    def test_raises_error_if_contents_but_no_snakebids_file(self, tmp_path: Path):
        (tmp_path / "some_file").touch()
        with pytest.raises(RunError):
            output._get_snakebids_file(tmp_path)

    def test_raises_error_if_malformed_snakebids_file_found(self, tmp_path: Path):
        (tmp_path / "some_file").touch()
        (tmp_path / ".snakebids").touch()
        with pytest.raises(RunError):
            output._get_snakebids_file(tmp_path)

    def test_returns_contents_of_snakebids_file_if_valid(self, tmp_path: Path):
        (tmp_path / "some_file").touch()
        (tmp_path / ".snakebids").touch()
        with (tmp_path / ".snakebids").open("w") as f:
            json.dump({"mode": "foo"}, f)
        path = output._get_snakebids_file(tmp_path)
        assert path == {"mode": "foo"}

    def test_returns_none_if_directory_empty(self, tmp_path: Path):
        assert output._get_snakebids_file(tmp_path) is None

    def test_returns_none_if_directory_nonexistant(self, tmp_path: Path):
        empty_dir = tmp_path / "empty"
        empty_dir.mkdir()
        assert output._get_snakebids_file(empty_dir) is None


class TestPrepareBidsappOutput:
    def test_creates_new_empty_bidsapp(self, fake_snakemake: Path):
        out = fake_snakemake / "new-output"
        output.prepare_bidsapp_output(out, False)

        assert out.exists()
        assert dirlen(out) == 1
        with (out / ".snakebids").open() as f:
            assert json.load(f)["mode"] == "bidsapp"

    def test_fails_when_directory_has_contents(self, fake_snakemake: Path):
        with pytest.raises(RunError):
            output.prepare_bidsapp_output(fake_snakemake / "randomdir", False)

    def test_success_on_directory_with_contents_when_forced(self, fake_snakemake: Path):
        out = fake_snakemake / "randomdir"
        output.prepare_bidsapp_output(out, True)
        assert out.exists()
        assert dirlen(out) == 2
        with (out / ".snakebids").open() as f:
            assert json.load(f)["mode"] == "bidsapp"



from __future__ import annotations

from collections.abc import Iterable, Iterator, Mapping
from typing import Literal, TypeVar, overload

import more_itertools as itx

from snakebids.types import ZipList, ZipListLike
from snakebids.utils.containers import ContainerBag, MultiSelectDict, RegexContainer

T_co = TypeVar("T_co", bound=list[str] | str, covariant=True)


@overload
def filter_list(
    zip_list: ZipListLike,
    filters: Mapping[str, Iterable[str] | str],
    return_indices_only: Literal[False] = ...,
    regex_search: bool = ...,
) -> ZipList: ...


@overload
def filter_list(
    zip_list: ZipListLike,
    filters: Mapping[str, Iterable[str] | str],
    return_indices_only: Literal[True],
    regex_search: bool = ...,
) -> list[int]: ...


def filter_list(
    zip_list: ZipListLike,
    filters: Mapping[str, Iterable[str] | str],
    return_indices_only: bool = False,
    regex_search: bool = False,
) -> ZipList | list[int]:
    """Filter zip_list, including only entries with provided entity values.

    Parameters
    ----------
    zip_list
        generated zip lists dict from config file to filter

    filters
        wildcard values to filter the zip lists

    return_indices_only
        return the indices of the matching wildcards

    regex_search
        Use regex matching to filter instead of the default equality check.


    Examples
    --------
    ::

        >>> import snakebids

    Filtering to get all ``subject='01'`` scans::

        >>> snakebids.filter_list(
        ...     {
        ...         'dir': ['AP','PA','AP','PA', 'AP','PA','AP','PA'],
        ...         'acq': ['98','98','98','98','99','99','99','99'],
        ...         'subject': ['01','01','02','02','01','01','02','02' ]
        ...     },
        ...     {'subject': '01'}
        ... ) == {
        ...     'dir': ['AP', 'PA', 'AP', 'PA'],
        ...     'acq': ['98', '98', '99', '99'],
        ...     'subject': ['01', '01', '01', '01']
        ... }
        True

    Filtering to get all ``acq='98'`` scans::

        >>> snakebids.filter_list(
        ...     {
        ...         'dir': ['AP','PA','AP','PA', 'AP','PA','AP','PA'],
        ...         'acq': ['98','98','98','98','99','99','99','99'],
        ...         'subject': ['01','01','02','02','01','01','02','02' ]
        ...     },
        ...     {'acq': '98'}
        ... ) == {
        ...     'dir': ['AP', 'PA', 'AP', 'PA'],
        ...     'acq': ['98', '98', '98', '98'],
        ...     'subject': ['01', '01', '02', '02']
        ... }
        True

    Filtering to get all ``dir=='AP'`` scans::

        >>> snakebids.filter_list(
        ...     {
        ...         'dir': ['AP','PA','AP','PA', 'AP','PA','AP','PA'],
        ...         'acq': ['98','98','98','98','99','99','99','99'],
        ...         'subject': ['01','01','02','02','01','01','02','02' ]
        ...     },
        ...     {'dir': 'AP'}
        ... ) == {
        ...     'dir': ['AP', 'AP', 'AP', 'AP'],
        ...     'acq': ['98', '98', '99', '99'],
        ...     'subject': ['01', '02', '01', '02']
        ... }
        True

    Filtering to get all ``subject='03'`` scans (i.e. no matches)::

        >>> snakebids.filter_list(
        ...     {
        ...         'dir': ['AP','PA','AP','PA', 'AP','PA','AP','PA'],
        ...         'acq': ['98','98','98','98','99','99','99','99'],
        ...         'subject': ['01','01','02','02','01','01','02','02' ]
        ...     },
        ...     {'subject': '03'}
        ... ) == {
        ...     'dir': [],
        ...     'acq': [],
        ...     'subject': []
        ... }
        True
    """
    # Save filters into memory as sets for quick access later
    if regex_search:
        filter_sets = {
            key: ContainerBag(*(RegexContainer(r) for r in itx.always_iterable(vals)))
            for key, vals in filters.items()
        }
    else:
        filter_sets = {
            key: set(itx.always_iterable(vals)) for key, vals in filters.items()
        }

    # Get a set {0,1,2,3...n-1} where n is the length of any one of the lists in
    # zip_list
    keep_indices = set(_get_zip_list_indices(zip_list)).intersection(
        *(
            {i for i, v in enumerate(zip_list[key]) if v in container}
            for key, container in filter_sets.items()
            if key in zip_list
        )
    )

    # Now we have the indices, so filter the lists
    if return_indices_only:
        return list(keep_indices)
    return MultiSelectDict(
        {key: [val[i] for i in keep_indices] for key, val in zip_list.items()}
    )


def get_filtered_ziplist_index(
    zip_list: ZipList,
    wildcards: dict[str, str],
    subj_wildcards: dict[str, str],
) -> int | list[int]:
    """Return the indices of all entries matching the filter query.

    Parameters
    ----------
    zip_list : dict
        lists for scans in a dataset, zipped to get each instance
    wildcards: dict
        wildcards for the single instance for querying it's index
    subj_wildcards: dict
        keys of this dictionary are used to pick out the subject/(session)
        from the wildcards

    Examples
    --------
    >>> import snakebids

    In this example, we have a dataset where with scans from two subjects,
    where each subject has ``dir-AP`` and ``dir-PA`` scans, along with
    ``acq-98`` and ``acq-99``:

    * ``sub-01_acq-98_dir-AP_dwi.nii.gz``
    * ``sub-01_acq-98_dir-PA_dwi.nii.gz``
    * ``sub-01_acq-99_dir-AP_dwi.nii.gz``
    * ``sub-01_acq-99_dir-PA_dwi.nii.gz``
    * ``sub-02_acq-98_dir-AP_dwi.nii.gz``
    * ``sub-02_acq-98_dir-PA_dwi.nii.gz``
    * ``sub-02_acq-99_dir-AP_dwi.nii.gz``
    * ``sub-02_acq-99_dir-PA_dwi.nii.gz``

    The ``zip_list`` produced by ``generate_inputs()`` is the set of entities
    that when zipped together, e.g. with ``expand(path, zip, **zip_list)``,
    produces the entity combinations that refer to each scan::

        {
            'dir': ['AP','PA','AP','PA', 'AP','PA','AP','PA'],
            'acq': ['98','98','98','98','99','99','99','99'],
            'subject': ['01','01','02','02','01','01','02','02']
        }

    The ``filter_list()`` function produces a subset of the entity
    combinations as a filtered zip list. This is used e.g. to get all the
    scans for a single subject.

    This ``get_filtered_ziplist_index()`` function performs ``filter_list()``
    twice:

    1. Using the ``subj_wildcards`` (e.g.: ``'subject': '{subject}'``) to get
       a subject/session-specific ``zip_list``.
    2. To return the indices from that list of the matching wildcards.

    In this example, if the wildcards parameter was::

        {'dir': 'PA', 'acq': '99', 'subject': '01'}

    Then the first (subject/session-specific) filtered list provides this zip
    list::

        {
            'dir': ['AP','PA','AP','PA'],
            'acq': ['98','98','99','99'],
            'subject': ['01','01','01','01']
        }

    which has 4 combinations, and thus are indexed from 0 to 3.

    The returned value would then be the index (or indices) that matches the
    wildcards. In this case, since the wildcards were
    ``{'dir': 'PA', 'acq': '99', 'subject':'01'}``, the return index is 3. ::

        >>> snakebids.get_filtered_ziplist_index(
        ...     {
        ...         'dir': ['AP','PA','AP','PA', 'AP','PA','AP','PA'],
        ...         'acq': ['98','98','98','98','99','99','99','99'],
        ...         'subject': ['01','01','02','02','01','01','02','02' ]
        ...     },
        ...     {'dir': 'PA', 'acq': '99', 'subject': '01'},
        ...     {'subject': '{subject}' }
        ... )
        3
    """
    # get the subject/(session) dict:
    subj_dict = {key: wildcards[key] for key in subj_wildcards}

    # now filter the list based on subj_wildcards
    zip_list_filtered = filter_list(zip_list, subj_dict)

    # get the index of the wildcard from this filtered list
    indices = filter_list(zip_list_filtered, wildcards, return_indices_only=True)
    if len(indices) == 1:
        return indices[0]
    return indices


def _get_zip_list_indices(zip_list: ZipListLike) -> Iterator[int]:
    """Convert a zip_list into its indices.

    Generates a sequence of numbers from 0 up to the length of the zip_lists. For
    example, the zip list:

    zip_list = {
        "subject": ["001", "002", "001", "002"],
        "contrast": ["T1w", "T1w", "T2w", "T2w"]
    }

    would lead to:

    (0, 1, 2, 3)

    Parameters
    ----------
    zip_list : dict[str, list[str]]
        Zip_list to be converted

    Yields
    ------
    integers
        The indices of the zip_list
    """
    if not zip_list:
        return

    # Retrieve the first zip_list (all zip_lists should be the same, so it doesn't
    # matter which one we take)
    sample_zip_list = zip_list[itx.first(zip_list)]

    # Return a sequence of numbers 0, 1, 2, 3, ... n-1 where n is the length of the
    # zip list
    yield from range(len(sample_zip_list))



"""
This type stub file was generated by pyright.
"""

from typing import Sequence

class BIDSMetadata(dict):
    """Metadata dictionary that reports the associated file on lookup failures."""

    def __init__(self, source_file) -> None: ...
    def __getitem__(self, key): ...

class PaddedInt(int):
    """Integer type that preserves zero-padding

    Acts like an int in almost all ways except that string formatting
    will keep the original zero-padding. Numeric format specifiers

    >>> PaddedInt(1)
    1
    >>> p2 = PaddedInt("02")
    >>> p2
    02
    >>> str(p2)
    '02'
    >>> p2 == 2
    True
    >>> p2 in range(3)
    True
    >>> f"{p2}"
    '02'
    >>> f"{p2:s}"
    '02'
    >>> f"{p2!s}"
    '02'
    >>> f"{p2!r}"
    '02'
    >>> f"{p2:d}"
    '2'
    >>> f"{p2:03d}"
    '002'
    >>> f"{p2:f}"
    '2.000000'
    >>> {2: "val"}.get(p2)
    'val'
    >>> {p2: "val"}.get(2)
    'val'

    Note that arithmetic will break the padding.

    >>> str(p2 + 1)
    '3'
    """

    def __init__(self, val) -> None: ...
    def __eq__(self, val) -> bool: ...
    def __str__(self) -> str: ...
    def __repr__(self): ...
    def __format__(self, format_spec):  # -> str:
        """Format a padded integer

        If a format spec can be used on a string, apply it to the zero-padded string.
        Otherwise format as an integer.
        """
        ...
    def __hash__(self) -> int: ...

def parse_file_entities(
    filename: str,
    entities: Sequence[str] = ...,
    config: str = ...,
    include_unmatched: bool = ...,
) -> dict[str, str]:
    """Parse the passed filename for entity/value pairs.

    Parameters
    ----------
    filename : str
        The filename to parse for entity values
    entities : list or None, optional
        An optional list of Entity instances to use in extraction.
        If passed, the config argument is ignored. Default is None.
    config : str or :obj:`bids.layout.models.Config` or list or None, optional
        One or more :obj:`bids.layout.models.Config` objects or names of
        configurations to use in matching. Each element must be a
        :obj:`bids.layout.models.Config` object, or a valid
        :obj:`bids.layout.models.Config` name (e.g., 'bids' or 'derivatives').
        If None, all available configs are used. Default is None.
    include_unmatched : bool, optional
        If True, unmatched entities are included in the returned dict,
        with values set to None.
        If False (default), unmatched entities are ignored.

    Returns
    -------
    dict
        Keys are Entity names and values are the values from the filename.
    """
    ...

def add_config_paths(**kwargs):  # -> None:
    """Add to the pool of available configuration files for BIDSLayout.

    Parameters
    ----------
    kwargs : dict
        Dictionary specifying where to find additional config files.
        Keys are names, values are paths to the corresponding .json file.

    Examples
    --------
    > add_config_paths(my_config='/path/to/config')
    > layout = BIDSLayout('/path/to/bids', config=['bids', 'my_config'])
    """
    ...



from __future__ import annotations

import warnings
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

import snakebids.paths._templates.spec_func as spec_template
from snakebids.paths import bids_factory, specs
from snakebids.paths._config import reset_bids_spec, set_bids_spec
from snakebids.paths._presets import bids
from snakebids.paths._utils import find_entity, get_spec_path, load_spec
from snakebids.paths.specs import v0_0_0


@pytest.fixture(autouse=True)
def _spec_reset():  # type: ignore
    reset_bids_spec()


def test_reset_bids_spec_clears_cache(mocker: MockerFixture):
    load_spec_spy = mocker.spy(specs, "load_spec")
    _ = specs.v0_0_0
    load_spec_spy.assert_called_once()
    load_spec_spy.reset_mock()
    _ = specs.v0_0_0
    load_spec_spy.assert_not_called()
    reset_bids_spec()
    load_spec_spy.reset_mock()
    _ = specs.v0_0_0
    load_spec_spy.assert_called_once()


def test_all_entries_define_entity():
    spec = v0_0_0()
    for item in spec:
        assert "entity" in item


def test_subject_dir_can_be_excluded():
    spec = v0_0_0(subject_dir=False)
    subject = find_entity(spec, "subject")
    assert subject.get("dir") is False


def test_session_dir_can_be_excluded():
    spec = v0_0_0(session_dir=False)
    session = find_entity(spec, "session")
    assert session.get("dir") is False


def test_spec_can_be_set_with_str():
    set_bids_spec("v0_0_0")
    assert bids(acquisition="foo") == "acquisition-foo"
    set_bids_spec("v0_11_0")
    assert bids(acquisition="foo") == "acq-foo"


def test_spec_can_be_set_with_obj():
    set_bids_spec(specs.v0_0_0())
    assert bids(acquisition="foo") == "acquisition-foo"
    set_bids_spec(specs.v0_11_0())
    assert bids(acquisition="foo") == "acq-foo"


def test_using_include_subject_dir_raises_warning():
    with pytest.warns(UserWarning, match="include_session_dir and include_subject_dir"):
        bids(subject="001", include_subject_dir=False)
    with pytest.warns(UserWarning, match="include_session_dir and include_subject_dir"):
        bids(session="001", include_session_dir=False)


def test_include_subject_dir_can_remove_dir():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        assert len(Path(bids(subject="001", include_subject_dir=False)).parents) == 1
        assert len(Path(bids(session="001", include_session_dir=False)).parents) == 1


class TestSpecDocstrings:
    def clean(self, item: str):
        return " ".join(item.split())

    def specfuncs(self):
        for attr in dir(specs):
            print(attr)
            if attr.startswith("v") or attr == "latest":
                yield getattr(specs, attr)

    def test_all_specs_have_docstrings(self):
        for spec in self.specfuncs():
            assert isinstance(spec.__doc__, str)

    def test_all_specs_have_format_example(self):
        for spec in self.specfuncs():
            assert "Formatted as::" in spec.__doc__

    def test_spec_has_correct_docstring(self):
        spec = load_spec(get_spec_path("v0_0_0"))
        assert (
            self.clean(specs.v0_0_0.__doc__).startswith(self.clean(spec["description"]))  # type: ignore
        )

    def test_spec_with_description_gets_default_docstring(self, mocker: MockerFixture):
        spec = load_spec(get_spec_path("v0_0_0"))
        _ = spec.pop("description", None)
        mocker.patch.object(specs, "load_spec", return_value=spec)
        assert self.clean(specs.v0_0_0.__doc__).startswith(  # type: ignore
            self.clean(spec_template.DEFAULT_DESCRIPTION.format(version="v0.0.0"))
        )

    def test_no_format_example_when_no_docstring_parser(self, mocker: MockerFixture):
        mocker.patch.object(
            spec_template, "_import_docstring_parser", side_effect=ImportError()
        )
        # Need to reset after mocking because the reset collects the latest spec
        reset_bids_spec()
        for spec in self.specfuncs():
            assert "Formatted as::" not in spec.__doc__, spec


class TestCustomEntityWarnings:
    def test_using_custom_entities_with_default_bids_raises_warning(self):
        with pytest.warns(UserWarning, match="spec has not been explicitly declared"):
            bids(foo="bar")

    def test_no_warning_when_spec_declared(self):
        set_bids_spec("v0_0_0")
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            bids(foo="bar")

    def test_no_warning_when_bids_explicitly_generated(self):
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            bids_factory(specs.v0_0_0())(foo="bar")

    def test_no_warning_in_interactive_mode(self, mocker: MockerFixture):
        mocker.patch(
            "snakebids.paths._factory.in_interactive_session", return_value=True
        )
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            bids(foo="bar")



# type: ignore
__submodules__ = ["_factory", "_presets", "_config", "_utils"]

# <AUTOGEN_INIT>
import lazy_loader

__getattr__, __dir__, __all__ = lazy_loader.attach_stub(__name__, __file__)

__all__ = [
    "BidsFunction",
    "BidsPathEntitySpec",
    "BidsPathSpec",
    "BidsPathSpecFile",
    "bids",
    "bids_factory",
    "set_bids_spec",
]
# </AUTOGEN_INIT>



"""
This type stub file was generated by pyright.
"""

from ruamel.yaml.error import MarkedYAMLError
from ruamel.yaml.tokens import *
from ruamel.yaml.events import *
from ruamel.yaml.tag import Tag

if False: ...
__all__ = ["Parser", "RoundTripParser", "ParserError"]

def xprintf(*args: Any, **kw: Any) -> Any: ...

class ParserError(MarkedYAMLError): ...

class Parser:
    DEFAULT_TAGS = ...
    def __init__(self, loader: Any) -> None: ...
    def reset_parser(self) -> None: ...
    def dispose(self) -> None: ...
    @property
    def scanner(self) -> Any: ...
    @property
    def resolver(self) -> Any: ...
    def check_event(self, *choices: Any) -> bool: ...
    def peek_event(self) -> Any: ...
    def get_event(self) -> Any: ...
    def parse_stream_start(self) -> Any: ...
    def parse_implicit_document_start(self) -> Any: ...
    def parse_document_start(self) -> Any: ...
    def parse_document_end(self) -> Any: ...
    def parse_document_content(self) -> Any: ...
    def process_directives(self) -> Any: ...
    def parse_block_node(self) -> Any: ...
    def parse_flow_node(self) -> Any: ...
    def parse_block_node_or_indentless_sequence(self) -> Any: ...
    def select_tag_transform(self, tag: Tag) -> None: ...
    def parse_node(self, block: bool = ..., indentless_sequence: bool = ...) -> Any: ...
    def parse_block_sequence_first_entry(self) -> Any: ...
    def parse_block_sequence_entry(self) -> Any: ...
    def parse_indentless_sequence_entry(self) -> Any: ...
    def parse_block_mapping_first_key(self) -> Any: ...
    def parse_block_mapping_key(self) -> Any: ...
    def parse_block_mapping_value(self) -> Any: ...
    def parse_flow_sequence_first_entry(self) -> Any: ...
    def parse_flow_sequence_entry(self, first: bool = ...) -> Any: ...
    def parse_flow_sequence_entry_mapping_key(self) -> Any: ...
    def parse_flow_sequence_entry_mapping_value(self) -> Any: ...
    def parse_flow_sequence_entry_mapping_end(self) -> Any: ...
    def parse_flow_mapping_first_key(self) -> Any: ...
    def parse_flow_mapping_key(self, first: Any = ...) -> Any: ...
    def parse_flow_mapping_value(self) -> Any: ...
    def parse_flow_mapping_empty_value(self) -> Any: ...
    def process_empty_scalar(self, mark: Any, comment: Any = ...) -> Any: ...
    def move_token_comment(
        self, token: Any, nt: Optional[Any] = ..., empty: Optional[bool] = ...
    ) -> Any: ...

class RoundTripParser(Parser):
    """roundtrip is a safe loader, that wants to see the unmangled tag"""
    def select_tag_transform(self, tag: Tag) -> None: ...
    def move_token_comment(
        self, token: Any, nt: Optional[Any] = ..., empty: Optional[bool] = ...
    ) -> Any: ...

class RoundTripParserSC(RoundTripParser):
    """roundtrip is a safe loader, that wants to see the unmangled tag"""
    def move_token_comment(
        self: Any, token: Any, nt: Any = ..., empty: Optional[bool] = ...
    ) -> None: ...
    def distribute_comment(self, comment: Any, line: Any) -> Any: ...



# ruff: noqa: PLR2004
from __future__ import annotations

import argparse
from typing import Any

import pytest
from pytest_mock import MockerFixture

from snakebids import bidsapp, plugins
from snakebids.bidsapp.args import ArgumentGroups


class TestRunner:
    def setup_method(self):
        self.hooks_run = 0

    @bidsapp.hookimpl
    def initialize_config(self, config: dict[str, Any]):
        self.hooks_run += 1
        assert not config
        config["initialized"] = True

    @bidsapp.hookimpl
    def add_cli_arguments(
        self,
        parser: argparse.ArgumentParser,
        config: dict[str, Any],
        argument_groups: ArgumentGroups,
    ):
        self.hooks_run += 1
        assert len(config) == 1
        assert config["initialized"]
        assert len(parser._actions) == 1
        assert len(argument_groups) == 2
        config["added_args"] = True
        parser.add_argument("arg_one")
        parser.add_argument("--arg-two")

    @bidsapp.hookimpl
    def get_argv(self, argv: list[str], config: dict[str, Any]):
        self.hooks_run += 1
        assert config["added_args"]
        config["provided_args"] = argv
        assert argv == ["default"]
        return ["known", "--arg-two", "known", "unknown"]

    @bidsapp.hookimpl
    def handle_unknown_args(self, args: list[str], config: dict[str, Any]):
        self.hooks_run += 1
        assert args == ["unknown"]
        assert len(config) == 3
        assert config["initialized"]
        assert config["added_args"]
        assert config["provided_args"] == ["default"]
        config["unknown_args"] = args

    @bidsapp.hookimpl
    def update_cli_namespace(self, namespace: dict[str, Any], config: dict[str, Any]):
        self.hooks_run += 1
        assert len(config) == 4
        assert config["initialized"]
        assert config["added_args"]
        assert config["unknown_args"] == ["unknown"]
        assert len(namespace) == 2
        assert namespace["arg_one"] == "known"
        assert namespace["arg_two"] == "known"
        config["arg_two"] = namespace.pop("arg_two")

    @bidsapp.hookimpl
    def finalize_config(self, config: dict[str, Any]):
        self.hooks_run += 1
        assert len(config) == 6
        assert config["initialized"]
        assert config["added_args"]
        assert config["unknown_args"] == ["unknown"]
        assert config["arg_one"] == "known"
        assert config["arg_two"] == "known"
        config["finalized"] = True

    @bidsapp.hookimpl
    def run(self, config: dict[str, Any]):
        self.hooks_run += 1
        assert len(config) == 7
        assert config["initialized"]
        assert config["added_args"]
        assert config["unknown_args"] == ["unknown"]
        assert config["arg_one"] == "known"
        assert config["arg_two"] == "known"
        assert config["finalized"]
        config["run"] = True

    def test_build_parser(self):
        app = bidsapp.app(plugins=[self])
        app.build_parser()
        assert self.hooks_run == 2
        app.build_parser()
        assert self.hooks_run == 2

    def test_parse_args(self):
        app = bidsapp.app(plugins=[self])
        app.parse_args(args=["default"])
        assert self.hooks_run == 6
        app.parse_args(args=["default"])
        assert self.hooks_run == 6

    def test_run(self):
        app = bidsapp.app(plugins=[self])
        assert app.run(args=["default"]) is None
        assert self.hooks_run == 7
        assert app.run(args=["default"]) is None
        assert self.hooks_run == 7


class TestNoArgvHook:
    @bidsapp.hookimpl
    def handle_unknown_args(self, args: list[str]):
        assert args == ["mocked", "args"]

    def test_with_mocked_args(self, mocker: MockerFixture):
        from snakebids.bidsapp.run import sys  # noqa: PLC0415

        mocker.patch.object(sys, "argv", ["program", "mocked", "args"])
        app = bidsapp.app(plugins=[self])
        app.parse_args()

    def test_with_provided_args(self):
        app = bidsapp.app(plugins=[self])
        app.parse_args(args=["mocked", "args"])


class TestDependencies:
    DEPENDENCIES = (plugins.BidsArgs(),)

    def setup_method(self):
        self.called = False

    @bidsapp.hookimpl
    def add_cli_arguments(
        self,
        parser: argparse.ArgumentParser,
    ):
        self.called = True
        assert len(parser._actions) > 1

    def test_dependencies_are_added(self):
        app = bidsapp.app(plugins=[self])
        assert len(app.pm.list_name_plugin()) == 2

    def test_dependences_registered_after(self):
        app = bidsapp.app(plugins=[self])
        app.build_parser()
        assert self.called

    def test_dependencies_wont_register_twice(self):
        app = bidsapp.app(
            plugins=[
                self,
                plugins.BidsArgs(
                    bids_dir=False,
                    output_dir=False,
                    analysis_level=False,
                    participant_label=False,
                    exclude_participant_label=False,
                    # leave derivatives in so that the DependenciesPlugin
                    # assertion doesn't get tripped
                ),
            ]
        )
        app.run()
        assert self.called
        assert len(app.parser._actions) == 2


class TestDependencyDeduplication:
    DEPENDENCIES = (
        plugins.BidsArgs(),
        plugins.ComponentEdit(),
        plugins.CliConfig(),
        plugins.Pybidsdb(),
        plugins.BidsValidator(),
    )

    def get_plugins(self):
        # We need to regenerate all instances again to check that the `isinstance`
        # equality checks in all plugins are working
        return (
            plugins.BidsArgs(),
            plugins.ComponentEdit(),
            plugins.CliConfig(),
            plugins.Pybidsdb(),
            plugins.BidsValidator(),
        )

    def test_duplicate_normally_caught(self):
        with pytest.raises(ValueError, match="Plugin already registered"):
            bidsapp.app(plugins=[*self.get_plugins(), *self.DEPENDENCIES])

    def test_no_duplicate_plugins_added_when_dependencies(self):
        plugins = self.get_plugins()
        app = bidsapp.app(plugins=[self, *plugins])
        assert len(app.pm.list_name_plugin()) == len(plugins) + 1


class TestArgumentGroups:
    def setup_method(self):
        self.group_names = None

    @bidsapp.hookimpl
    def add_cli_arguments(
        self,
        argument_groups: ArgumentGroups,
    ):
        assert argument_groups["new"]
        self.group_names = list(argument_groups)

    def test_argument_groups_visible_to_add_arg_hook(self):
        app = bidsapp.app(plugins=[self])
        app.parser.add_argument_group("foo")
        app.run()
        assert self.group_names
        assert set(self.group_names).issuperset({"foo", "new"})
        assert app.argument_groups["new"].title == "new"



from __future__ import annotations

import argparse
import sys
from collections.abc import Callable, Iterable
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeAlias,
    TypedDict,
    TypeVar,
    overload,
)

from typing_extensions import Required, Unpack

if TYPE_CHECKING:
    from argparse import _ArgumentGroup

_T = TypeVar("_T")


if sys.version_info >= (3, 11):

    class AddArgumentArgs(TypedDict, Generic[_T], total=False):
        """Arguments for add_argument, with ``dest`` required."""

        action: str | type[argparse.Action]
        nargs: int | str | None
        const: Any
        default: Any
        type: Callable[[Any], _T] | argparse.FileType
        choices: Iterable[_T] | None
        required: bool
        help: str | None
        metavar: str | tuple[str, ...] | None
        dest: Required[str]
        version: str

    AnyArgumentArgs: TypeAlias = "AddArgumentArgs[Any]"

else:

    class AddArgumentArgs(TypedDict, total=False):
        """Arguments for add_argument, with ``dest`` required."""

        action: str | type[argparse.Action]
        nargs: int | str | None
        const: Any
        default: Any
        type: Callable[[Any], Any] | argparse.FileType
        choices: Iterable[Any] | None
        required: bool
        help: str | None
        metavar: str | tuple[str, ...] | None
        dest: Required[str]
        version: str

    AnyArgumentArgs: TypeAlias = "AddArgumentArgs"


class PluginBase:
    """Optional base class for snakebids plugins providing utility methods."""

    PREFIX = ""

    @overload
    def pop(self, mapping: dict[str, Any], key: str, default: Any, /) -> Any: ...

    @overload
    def pop(self, mapping: dict[str, Any], key: str, /) -> Any: ...

    def pop(self, mapping: dict[str, Any], *args: Any):
        """Remove specified key from mapping, prepending the plugin prefix."""
        if not args:
            msg = "pop expected at least two arguments, got 1"
            raise TypeError(msg)
        if self.PREFIX:
            return mapping.pop(f"{self.PREFIX}.{args[0]}", *args[1:])
        return mapping.pop(*args)

    def get(self, mapping: dict[str, Any], key: str):
        """Retrieve specified key from mapping, prepending the plugin prefix."""
        if self.PREFIX:
            return mapping.get(f"{self.PREFIX}.{key}")
        return mapping.get(key)

    def assign(self, mapping: dict[str, Any], key: str, value: Any):
        """Assign specified key from mapping, prepending the plugin prefix."""
        if self.PREFIX:
            mapping[f"{self.PREFIX}.{key}"] = value
        else:
            mapping[key] = value

    def try_add_argument(
        self,
        parser: argparse.ArgumentParser | _ArgumentGroup,
        *name_or_flags: str,
        **kwargs: Unpack[AnyArgumentArgs],
    ) -> argparse.Action | None:
        """Add argument to parser if provided dest is not already present."""
        if self.PREFIX and not kwargs["dest"].startswith(self.PREFIX):
            kwargs["dest"] = f"{self.PREFIX}.{kwargs['dest']}"
        if kwargs["dest"] in {a.dest for a in parser._actions}:  # noqa: SLF001
            return None
        return parser.add_argument(*name_or_flags, **kwargs)

    def add_argument(
        self,
        parser: argparse.ArgumentParser | _ArgumentGroup,
        *name_or_flags: str,
        **kwargs: Unpack[AnyArgumentArgs],
    ) -> argparse.Action | None:
        """Add argument to parser, applying prefix to dest as necessary."""
        if self.PREFIX and not kwargs["dest"].startswith(self.PREFIX):
            kwargs["dest"] = f"{self.PREFIX}.{kwargs['dest']}"
        return parser.add_argument(*name_or_flags, **kwargs)



"""
This type stub file was generated by pyright.
"""

"""
    inflect.py: correctly generate plurals, ordinals, indefinite articles;
                convert numbers to words
    Copyright (C) 2010 Paul Dyson

    Based upon the Perl module Lingua::EN::Inflect by Damian Conway.

    The original Perl module Lingua::EN::Inflect by Damian Conway is
    available from http://search.cpan.org/~dconway/

    This module can be downloaded at http://pypi.org/project/inflect

methods:
          classical inflect
          plural plural_noun plural_verb plural_adj singular_noun no num a an
          compare compare_nouns compare_verbs compare_adjs
          present_participle
          ordinal
          number_to_words
          join
          defnoun defverb defadj defa defan

    INFLECTIONS:    classical inflect
          plural plural_noun plural_verb plural_adj singular_noun compare
          no num a an present_participle

    PLURALS:   classical inflect
          plural plural_noun plural_verb plural_adj singular_noun no num
          compare compare_nouns compare_verbs compare_adjs

    COMPARISONS:    classical
          compare compare_nouns compare_verbs compare_adjs

    ARTICLES:   classical inflect num a an

    NUMERICAL:      ordinal number_to_words

    USER_DEFINED:   defnoun defverb defadj defa defan

Exceptions:
 UnknownClassicalModeError
 BadNumValueError
 BadChunkingOptionError
 NumOutOfRangeError
 BadUserDefinedPatternError
 BadRcFileError
 BadGenderError

"""

class UnknownClassicalModeError(Exception): ...
class BadNumValueError(Exception): ...
class BadChunkingOptionError(Exception): ...
class NumOutOfRangeError(Exception): ...
class BadUserDefinedPatternError(Exception): ...
class BadRcFileError(Exception): ...
class BadGenderError(Exception): ...

__version__ = ...
STDOUT_ON = ...

def print3(txt): ...
def enclose(s): ...
def joinstem(cutpoint=..., words=...):  # -> str:
    """
    join stem of each word in words into a string for regex
    each word is truncated at cutpoint
    cutpoint is usually negative indicating the number of letters to remove
    from the end of each word

    e.g.
    joinstem(-2, ["ephemeris", "iris", ".*itis"]) returns
    (?:ephemer|ir|.*it)

    """
    ...

def bysize(words):  # -> dict[Unknown, Unknown]:
    """
    take a list of words and return a dict of sets sorted by word length
    e.g.
    ret[3]=set(['ant', 'cat', 'dog', 'pig'])
    ret[4]=set(['frog', 'goat'])
    ret[5]=set(['horse'])
    ret[8]=set(['elephant'])
    """
    ...

def make_pl_si_lists(
    lst, plending, siendingsize, dojoinstem=...
):  # -> tuple[list[Unknown], dict[Unknown, Unknown], dict[Unknown, Unknown], str] | tuple[list[Unknown], dict[Unknown, Unknown], dict[Unknown, Unknown]]:
    """
    given a list of singular words: lst
    an ending to append to make the plural: plending
    the number of characters to remove from the singular
        before appending plending: siendingsize
    a flag whether to create a joinstem: dojoinstem

    return:
    a list of pluralised words: si_list (called si because this is what you need to
                                         look for to make the singular)
    the pluralised words as a dict of sets sorted by word length: si_bysize
    the singular words as a dict of sets sorted by word length: pl_bysize
    if dojoinstem is True: a regular expression that matches any of the stems: stem
    """
    ...

pl_sb_irregular_s = ...
pl_sb_irregular = ...
pl_sb_irregular_caps = ...
pl_sb_irregular_compound = ...
si_sb_irregular = ...
keys = ...
si_sb_irregular_caps = ...
si_sb_irregular_compound = ...
keys = ...
pl_sb_z_zes_list = ...
pl_sb_z_zes_bysize = ...
pl_sb_ze_zes_list = ...
pl_sb_ze_zes_bysize = ...
pl_sb_C_is_ides_complete = ...
pl_sb_C_is_ides_endings = ...
pl_sb_C_is_ides = ...
pl_sb_C_is_ides_list = ...
pl_sb_C_a_ata_list = ...
pl_sb_U_a_ae_list = ...
pl_sb_C_a_ae_list = ...
pl_sb_C_en_ina_list = ...
pl_sb_U_um_a_list = ...
pl_sb_C_um_a_list = ...
pl_sb_U_us_i_list = ...
pl_sb_C_us_i_list = ...
pl_sb_C_us_us = ...
pl_sb_C_us_us_bysize = ...
pl_sb_U_on_a_list = ...
pl_sb_C_on_a_list = ...
pl_sb_C_o_i = ...
pl_sb_C_o_i_bysize = ...
si_sb_C_o_i_bysize = ...
pl_sb_C_o_i_stems = ...
pl_sb_U_o_os_complete = ...
si_sb_U_o_os_complete = ...
pl_sb_U_o_os_endings = ...
pl_sb_U_o_os_bysize = ...
si_sb_U_o_os_bysize = ...
pl_sb_U_ch_chs_list = ...
pl_sb_U_ex_ices_list = ...
pl_sb_U_ix_ices_list = ...
pl_sb_C_ex_ices_list = ...
pl_sb_C_ix_ices_list = ...
pl_sb_C_i_list = ...
pl_sb_C_im_list = ...
pl_sb_U_man_mans_list = ...
pl_sb_U_man_mans_caps_list = ...
pl_sb_uninflected_s_complete = ...
pl_sb_uninflected_s_endings = ...
pl_sb_uninflected_s = ...
pl_sb_uninflected_herd = ...
pl_sb_uninflected_complete = ...
pl_sb_uninflected_caps = ...
pl_sb_uninflected_endings = ...
pl_sb_uninflected_bysize = ...
pl_sb_singular_s_complete = ...
pl_sb_singular_s_endings = ...
pl_sb_singular_s_bysize = ...
si_sb_singular_s_complete = ...
si_sb_singular_s_endings = ...
si_sb_singular_s_bysize = ...
pl_sb_singular_s_es = ...
pl_sb_singular_s = ...
si_sb_ois_oi_case = ...
si_sb_uses_use_case = ...
si_sb_uses_use = ...
si_sb_ies_ie_case = ...
si_sb_ies_ie = ...
si_sb_oes_oe_case = ...
si_sb_oes_oe = ...
si_sb_z_zes = ...
si_sb_zzes_zz = ...
si_sb_ches_che_case = ...
si_sb_ches_che = ...
si_sb_xes_xe = ...
si_sb_sses_sse_case = ...
si_sb_sses_sse = ...
si_sb_ves_ve_case = ...
si_sb_ves_ve = ...
plverb_special_s = ...
pl_sb_postfix_adj = ...
pl_sb_postfix_adj_stems = ...
si_sb_es_is = ...
pl_prep_list = ...
pl_prep_list_da = ...
pl_prep_bysize = ...
pl_prep = ...
pl_sb_prep_dual_compound = ...
singular_pronoun_genders = ...
pl_pron_nom = ...
si_pron = ...
pl_pron_acc = ...
pl_pron_acc_keys = ...
pl_pron_acc_keys_bysize = ...
si_pron_acc_keys = ...
si_pron_acc_keys_bysize = ...

def get_si_pron(thecase, word, gender): ...

plverb_irregular_pres = ...
plverb_ambiguous_pres = ...
plverb_ambiguous_pres_keys = ...
plverb_irregular_non_pres = ...
plverb_ambiguous_non_pres = ...
pl_v_oes_oe = ...
pl_v_oes_oe_endings_size4 = ...
pl_v_oes_oe_endings_size5 = ...
pl_count_zero = ...
pl_count_one = ...
pl_adj_special = ...
pl_adj_special_keys = ...
pl_adj_poss = ...
pl_adj_poss_keys = ...
A_abbrev = ...
A_y_cons = ...
A_explicit_a = ...
A_explicit_an = ...
A_ordinal_an = ...
A_ordinal_a = ...
nth = ...
ordinal = ...
ordinal_suff = ...
unit = ...
teen = ...
ten = ...
mill = ...
def_classical = ...
all_classical = ...
no_classical = ...
string_to_constant = ...

class engine:
    def __init__(self) -> None: ...

    deprecated_methods = ...
    def __getattr__(self, meth): ...
    def defnoun(self, singular, plural):  # -> Literal[1]:
        """
        Set the noun plural of singular to plural.

        """
        ...
    def defverb(self, s1, p1, s2, p2, s3, p3):  # -> Literal[1]:
        """
        Set the verb plurals for s1, s2 and s3 to p1, p2 and p3 respectively.

        Where 1, 2 and 3 represent the 1st, 2nd and 3rd person forms of the verb.

        """
        ...
    def defadj(self, singular, plural):  # -> Literal[1]:
        """
        Set the adjective plural of singular to plural.

        """
        ...
    def defa(self, pattern):  # -> Literal[1]:
        """
        Define the indefinite article as 'a' for words matching pattern.

        """
        ...
    def defan(self, pattern):  # -> Literal[1]:
        """
        Define the indefinite article as 'an' for words matching pattern.

        """
        ...
    def checkpat(self, pattern):  # -> None:
        """
        check for errors in a regex pattern
        """
        ...
    def checkpatplural(self, pattern):  # -> None:
        """
        check for errors in a regex replace pattern
        """
        ...
    def ud_match(self, word, wordlist): ...
    def classical(self, **kwargs):  # -> None:
        """
        turn classical mode on and off for various categories

        turn on all classical modes:
        classical()
        classical(all=True)

        turn on or off specific claassical modes:
        e.g.
        classical(herd=True)
        classical(names=False)

        By default all classical modes are off except names.

        unknown value in args or key in kwargs raises
        exception: UnknownClasicalModeError

        """
        ...
    def num(self, count=..., show=...):  # -> str:
        """
        Set the number to be used in other method calls.

        Returns count.

        Set show to False to return '' instead.

        """
        ...
    def gender(self, gender):  # -> None:
        """
        set the gender for the singular of plural pronouns

        can be one of:
        'neuter'                ('they' -> 'it')
        'feminine'              ('they' -> 'she')
        'masculine'             ('they' -> 'he')
        'gender-neutral'        ('they' -> 'they')
        'feminine or masculine' ('they' -> 'she or he')
        'masculine or feminine' ('they' -> 'he or she')
        """
        ...
    def inflect(self, text):
        """
        Perform inflections in a string.

        e.g. inflect('The plural of cat is plural(cat)') returns
        'The plural of cat is cats'

        can use plural, plural_noun, plural_verb, plural_adj,
        singular_noun, a, an, no, ordinal, number_to_words,
        and prespart

        """
        ...
    def postprocess(self, orig, inflected): ...
    def partition_word(self, text): ...
    def plural(self, text, count=...):  # -> str:
        """
        Return the plural of text.

        If count supplied, then return text if count is one of:
            1, a, an, one, each, every, this, that
        otherwise return the plural.

        Whitespace at the start and end is preserved.

        """
        ...
    def plural_noun(self, text, count=...):  # -> str:
        """
        Return the plural of text, where text is a noun.

        If count supplied, then return text if count is one of:
            1, a, an, one, each, every, this, that
        otherwise return the plural.

        Whitespace at the start and end is preserved.

        """
        ...
    def plural_verb(self, text, count=...):  # -> str:
        """
        Return the plural of text, where text is a verb.

        If count supplied, then return text if count is one of:
            1, a, an, one, each, every, this, that
        otherwise return the plural.

        Whitespace at the start and end is preserved.

        """
        ...
    def plural_adj(self, text, count=...):  # -> str:
        """
        Return the plural of text, where text is an adjective.

        If count supplied, then return text if count is one of:
            1, a, an, one, each, every, this, that
        otherwise return the plural.

        Whitespace at the start and end is preserved.

        """
        ...
    def compare(self, word1, word2):  # -> Literal['eq', 'p:s', 's:p', 'p:p', False]:
        """
        compare word1 and word2 for equality regardless of plurality

        return values:
        eq - the strings are equal
        p:s - word1 is the plural of word2
        s:p - word2 is the plural of word1
        p:p - word1 and word2 are two different plural forms of the one word
        False - otherwise

        """
        ...
    def compare_nouns(
        self, word1, word2
    ):  # -> Literal['eq', 'p:s', 's:p', 'p:p', False]:
        """
        compare word1 and word2 for equality regardless of plurality
        word1 and word2 are to be treated as nouns

        return values:
        eq - the strings are equal
        p:s - word1 is the plural of word2
        s:p - word2 is the plural of word1
        p:p - word1 and word2 are two different plural forms of the one word
        False - otherwise

        """
        ...
    def compare_verbs(
        self, word1, word2
    ):  # -> Literal['eq', 'p:s', 's:p', 'p:p', False]:
        """
        compare word1 and word2 for equality regardless of plurality
        word1 and word2 are to be treated as verbs

        return values:
        eq - the strings are equal
        p:s - word1 is the plural of word2
        s:p - word2 is the plural of word1
        p:p - word1 and word2 are two different plural forms of the one word
        False - otherwise

        """
        ...
    def compare_adjs(
        self, word1, word2
    ):  # -> Literal['eq', 'p:s', 's:p', 'p:p', False]:
        """
        compare word1 and word2 for equality regardless of plurality
        word1 and word2 are to be treated as adjectives

        return values:
        eq - the strings are equal
        p:s - word1 is the plural of word2
        s:p - word2 is the plural of word1
        p:p - word1 and word2 are two different plural forms of the one word
        False - otherwise

        """
        ...
    def singular_noun(self, text, count=..., gender=...):  # -> str | Literal[False]:
        """
        Return the singular of text, where text is a plural noun.

        If count supplied, then return the singular if count is one of:
            1, a, an, one, each, every, this, that or if count is None
        otherwise return text unchanged.

        Whitespace at the start and end is preserved.

        """
        ...
    def get_count(self, count=...): ...
    def a(self, text, count=...):  # -> str:
        """
        Return the appropriate indefinite article followed by text.

        The indefinite article is either 'a' or 'an'.

        If count is not one, then return count followed by text
        instead of 'a' or 'an'.

        Whitespace at the start and end is preserved.

        """
        ...
    an = ...
    def no(self, text, count=...):  # -> str:
        """
        If count is 0, no, zero or nil, return 'no' followed by the plural
        of text.

        If count is one of:
            1, a, an, one, each, every, this, that
        return count followed by text.

        Otherwise return count follow by the plural of text.

        In the return value count is always followed by a space.

        Whitespace at the start and end is preserved.

        """
        ...
    def present_participle(self, word):  # -> str | LiteralString:
        """
        Return the present participle for word.

        word is the 3rd person singular verb.

        """
        ...
    def ordinal(self, num):  # -> str:
        """
        Return the ordinal of num.

        num can be an integer or text

        e.g. ordinal(1) returns '1st'
        ordinal('one') returns 'first'

        """
        ...
    def millfn(self, ind=...): ...
    def unitfn(self, units, mindex=...): ...
    def tenfn(self, tens, units, mindex=...): ...
    def hundfn(self, hundreds, tens, units, mindex): ...
    def group1sub(self, mo): ...
    def group1bsub(self, mo): ...
    def group2sub(self, mo): ...
    def group3sub(self, mo): ...
    def hundsub(self, mo): ...
    def tensub(self, mo): ...
    def unitsub(self, mo): ...
    def enword(self, num, group): ...
    def blankfn(self, mo):  # -> Literal['']:
        """do a global blank replace
        TODO: surely this can be done with an option to re.sub
              rather than this fn
        """
        ...
    def commafn(self, mo):  # -> Literal[',']:
        """do a global ',' replace
        TODO: surely this can be done with an option to re.sub
              rather than this fn
        """
        ...
    def spacefn(self, mo):  # -> Literal[' ']:
        """do a global ' ' replace
        TODO: surely this can be done with an option to re.sub
              rather than this fn
        """
        ...
    def number_to_words(
        self,
        num,
        wantlist=...,
        group=...,
        comma=...,
        andword=...,
        zero=...,
        one=...,
        decimal=...,
        threshold=...,
    ):
        """
        Return a number in words.

        group = 1, 2 or 3 to group numbers before turning into words
        comma: define comma
        andword: word for 'and'. Can be set to ''.
            e.g. "one hundred and one" vs "one hundred one"
        zero: word for '0'
        one: word for '1'
        decimal: word for decimal point
        threshold: numbers above threshold not turned into words

        parameters not remembered from last call. Departure from Perl version.
        """
        ...
    def join(
        self, words, sep=..., sep_spaced=..., final_sep=..., conj=..., conj_spaced=...
    ):  # -> str:
        """
        Join words into a list.

        e.g. join(['ant', 'bee', 'fly']) returns 'ant, bee, and fly'

        options:
        conj: replacement for 'and'
        sep: separator. default ',', unless ',' is in the list then ';'
        final_sep: final separator. default ',', unless ',' is in the list then ';'
        conj_spaced: boolean. Should conj have spaces around it

        """
        ...



"""
This type stub file was generated by pyright.
"""

if False: ...
__all__ = [
    "ScalarString",
    "LiteralScalarString",
    "FoldedScalarString",
    "SingleQuotedScalarString",
    "DoubleQuotedScalarString",
    "PlainScalarString",
    "PreservedScalarString",
]

class ScalarString(str):
    __slots__ = ...
    def __new__(cls, *args: Any, **kw: Any) -> Any: ...
    def replace(self, old: Any, new: Any, maxreplace: SupportsIndex = ...) -> Any: ...
    @property
    def anchor(self) -> Any: ...
    def yaml_anchor(self, any: bool = ...) -> Any: ...
    def yaml_set_anchor(self, value: Any, always_dump: bool = ...) -> None: ...

class LiteralScalarString(ScalarString):
    __slots__ = ...
    style = ...
    def __new__(cls, value: Text, anchor: Any = ...) -> Any: ...

PreservedScalarString = LiteralScalarString

class FoldedScalarString(ScalarString):
    __slots__ = ...
    style = ...
    def __new__(cls, value: Text, anchor: Any = ...) -> Any: ...

class SingleQuotedScalarString(ScalarString):
    __slots__ = ...
    style = ...
    def __new__(cls, value: Text, anchor: Any = ...) -> Any: ...

class DoubleQuotedScalarString(ScalarString):
    __slots__ = ...
    style = ...
    def __new__(cls, value: Text, anchor: Any = ...) -> Any: ...

class PlainScalarString(ScalarString):
    __slots__ = ...
    style = ...
    def __new__(cls, value: Text, anchor: Any = ...) -> Any: ...

def preserve_literal(s: Text) -> Text: ...
def walk_tree(base: Any, map: Any = ...) -> None:
    """
    the routine here walks over a simple yaml tree (recursing in
    dict values and list items) and converts strings that
    have multiple lines to literal scalars

    You can also provide an explicit (ordered) mapping for multiple transforms
    (first of which is executed):
        map = ruamel.yaml.compat.ordereddict
        map['\n'] = preserve_literal
        map[':'] = SingleQuotedScalarString
        walk_tree(data, map=map)
    """
    ...



from .core import (
    BidsComponent,
    BidsComponentRow,
    BidsDataset,
    BidsDatasetDict,
    BidsPartialComponent,
    filter_list,
    generate_inputs,
    get_filtered_ziplist_index,
    get_wildcard_constraints,
    write_derivative_json,
)
from .paths import (
    BidsFunction,
    BidsPathEntitySpec,
    BidsPathSpec,
    BidsPathSpecFile,
    bids,
    bids_factory,
    set_bids_spec,
)

__all__ = [
    "BidsComponent",
    "BidsComponentRow",
    "BidsDataset",
    "BidsDatasetDict",
    "BidsFunction",
    "BidsPartialComponent",
    "BidsPathEntitySpec",
    "BidsPathSpec",
    "BidsPathSpecFile",
    "bids",
    "bids_factory",
    "filter_list",
    "generate_inputs",
    "get_filtered_ziplist_index",
    "get_wildcard_constraints",
    "set_bids_spec",
    "write_derivative_json",
]



"""
This type stub file was generated by pyright.
"""

from functools import lru_cache

"""
Database-related functionality.
"""

def get_database_file(path): ...

class ConnectionManager:
    def __init__(
        self, database_path=..., reset_database=..., config=..., init_args=...
    ) -> None: ...
    @classmethod
    def exists(cls, database_path): ...
    def reset_database(self, init_args, config=...): ...
    def save_database(
        self, database_path, replace_connection=...
    ):  # -> ConnectionManager | Self@ConnectionManager:
        """Save the current index as a SQLite3 DB at the specified location.

        Note: This is only necessary if a database_path was not specified
        at initialization, and the user now wants to save the index.
        If a database_path was specified originally, there is no need to
        re-save using this method.

        Parameters
        ----------
        database_path : str
            The path to the desired database folder. By default,
            uses .db_cache. If a relative path is passed, it is assumed to
            be relative to the BIDSLayout root directory.
        replace_connection : bool, optional
            If True, returns a new ConnectionManager that points to the newly
            created database. If False, returns the current instance.
        """
        ...
    @property
    def session(self): ...
    @property
    @lru_cache()
    def layout_info(self): ...
    def reset_session(self):  # -> None:
        """Force a new session."""
        ...



"""
This type stub file was generated by pyright.
"""
from __future__ import annotations

from functools import lru_cache

""" Model classes used in BIDSLayouts. """
Base = ...

class LayoutInfo(Base):
    """Contains information about a BIDSLayout's initialization parameters."""

    __tablename__ = ...
    root = ...
    absolute_paths = ...
    _derivatives = ...
    _config = ...
    def __init__(self, **kwargs) -> None: ...
    def __repr__(self): ...

class Config(Base):
    """Container for BIDS configuration information.

    Parameters
    ----------
    name : str
        The name to give the Config (e.g., 'bids').
    entities : list
        A list of dictionaries containing entity configuration
        information.
    default_path_patterns : list
        Optional list of patterns used to build new paths.
    session : :obj:`sqlalchemy.orm.session.Session` or None
        An optional SQLAlchemy session. If passed,
        the session is used to update the database with any newly created
        Entity objects. If None, no database update occurs.
    """

    __tablename__ = ...
    name = ...
    _default_path_patterns = ...
    entities: dict[str, Entity] = ...
    def __init__(
        self, name, entities=..., default_path_patterns=..., session=...
    ) -> None: ...
    @classmethod
    def load(self, config: str | dict[str, str], session=...) -> Config:
        """Load a Config instance from the passed configuration data.

        Parameters
        ----------
        config : str or dict
            A string or dict containing configuration information.
            Must be one of:
            * A string giving the name of a predefined config file
                (e.g., 'bids' or 'derivatives')
            * A path to a JSON file containing config information
            * A dictionary containing config information
        session : :obj:`sqlalchemy.orm.session.Session` or None
            An optional SQLAlchemy Session instance.
            If passed, the session is used to check the database for (and
            return) an existing Config with name defined in config['name'].

        Returns
        -------
        A Config instance.
        """
        ...
    def __repr__(self): ...

class BIDSFile(Base):
    """Represents a single file or directory in a BIDS dataset.

    Parameters
    ----------
    filename : str
        The path to the corresponding file.
    """

    __tablename__ = ...
    path: str = ...
    filename = ...
    dirname = ...
    entities: list[Entity] = ...
    is_dir = ...
    class_ = ...
    _associations = ...
    __mapper_args__ = ...
    def __init__(self, filename) -> None: ...
    def __getattr__(self, attr): ...
    def __repr__(self): ...
    def __fspath__(self): ...
    @property
    @lru_cache()
    def relpath(self):  # -> str:
        """Return path relative to layout root"""
        ...
    def get_associations(
        self, kind=..., include_parents=...
    ):  # -> Any | list[Unknown]:
        """Get associated files, optionally limiting by association kind.

        Parameters
        ----------
        kind : str
            The kind of association to return (e.g., "Child").
            By default, all associations are returned.
        include_parents : bool
            If True, files related through inheritance
            are included in the returned list. If False, only directly
            associated files are returned. For example, a file's JSON
            sidecar will always be returned, but other JSON files from
            which the sidecar inherits will only be returned if
            include_parents=True.

        Returns
        -------
        list
            A list of BIDSFile instances.
        """
        ...
    def get_metadata(self):  # -> BIDSMetadata:
        """Return all metadata associated with the current file."""
        ...
    def get_entities(self, metadata=..., values=...):  # -> dict[Unknown, Unknown]:
        """Return entity information for the current file.

        Parameters
        ----------
        metadata : bool or None
            If False (default), only entities defined
            for filenames (and not those found in the JSON sidecar) are
            returned. If True, only entities found in metadata files (and not
            defined for filenames) are returned. If None, all available
            entities are returned.
        values : str
            The kind of object to return in the dict's values.
            Must be one of:
                * 'tags': Returns only the tagged value--e.g., if the key
                is "subject", the value might be "01".
                * 'objects': Returns the corresponding Entity instance.

        Returns
        -------
        dict
            A dict, where keys are entity names and values are Entity
            instances.
        """
        ...
    def copy(
        self, path_patterns, symbolic_link=..., root=..., conflicts=...
    ):  # -> None:
        """Copy the contents of a file to a new location.

        Parameters
        ----------
        path_patterns : list
            List of patterns used to construct the new
            filename. See :obj:`build_path` documentation for details.
        symbolic_link : bool
            If True, use a symbolic link to point to the
            existing file. If False, creates a new file.
        root : str
            Optional path to prepend to the constructed filename.
        conflicts : str
            Defines the desired action when the output path already exists.
            Must be one of:
                'fail': raises an exception
                'skip' does nothing
                'overwrite': overwrites the existing file
                'append': adds  a suffix to each file copy, starting with 1
        """
        ...

class BIDSDataFile(BIDSFile):
    """Represents a single data file in a BIDS dataset.

    Derived from `BIDSFile` and provides additional functionality such as
    obtaining pandas DataFrame data representation (via `get_df`).
    """

    __mapper_args__ = ...
    def get_df(
        self, include_timing=..., adjust_onset=..., enforce_dtypes=..., **pd_args
    ):
        """Return the contents of a tsv file as a pandas DataFrame.

        Parameters
        ----------
        include_timing : bool
            If True, adds an "onset" column to dense
            timeseries files (e.g., *_physio.tsv.gz).
        adjust_onset : bool
            If True, the onset of each sample in a dense
            timeseries file is shifted to reflect the "StartTime" value in
            the JSON sidecar. If False, the first sample starts at 0 secs.
            Ignored if include_timing=False.
        enforce_dtypes : bool
            If True, enforces the data types defined in
            the BIDS spec on core columns (e.g., subject_id and session_id
            must be represented as strings).
        pd_args : dict
            Optional keyword arguments to pass onto pd.read_csv().

        Returns
        -------
        :obj:`pandas.DataFrame`
            A pandas DataFrame.
        """
        ...

class BIDSImageFile(BIDSFile):
    """Represents a single neuroimaging data file in a BIDS dataset.

    Derived from `BIDSFile` and provides additional functionality such as
    obtaining nibabel's image file representation (via `get_image`).
    """

    __mapper_args__ = ...
    def get_image(self, **kwargs):
        """Return the associated image file (if it exists) as a NiBabel object

        Any keyword arguments are passed to ``nibabel.load``.
        """
        ...

class BIDSJSONFile(BIDSFile):
    """Represents a single JSON metadata file in a BIDS dataset.

    Derived from `BIDSFile` and provides additional functionality for reading
    the contents of JSON files as either dicts or strings.
    """

    __mapper_args__ = ...
    def get_dict(self):  # -> dict[Unknown, Unknown]:
        """Return the contents of the current file as a dictionary."""
        ...
    def get_json(self):
        """Return the contents of the current file as a JSON string."""
        ...

class Entity(Base):
    """
    Represents a single entity defined in the JSON config.

    Parameters
    ----------
    name : str
        The name of the entity (e.g., 'subject', 'run', etc.)
    pattern : str
        A regex pattern used to match against file names.
        Must define at least one group, and only the first group is
        kept as the match.
    mandatory : bool
        If True, every File _must_ match this entity.
    directory : str
        Optional pattern defining a directory associated
        with the entity.
    dtype : str
        The optional data type of the Entity values. Must be
        one of 'int', 'float', 'bool', or 'str'. If None, no type
        enforcement will be attempted, which means the dtype of the
        value may be unpredictable.
    """

    __tablename__ = ...
    name = ...
    mandatory = ...
    pattern = ...
    directory = ...
    _dtype = ...
    files = ...
    regex: str = ...
    def __init__(
        self, name, pattern=..., mandatory=..., directory=..., dtype=...
    ) -> None: ...
    def __repr__(self): ...
    def __iter__(self): ...
    def __deepcopy__(self, memo): ...
    def match_file(self, f):  # -> PaddedInt | Any | None:
        """
        Determine whether the passed file matches the Entity.

        Parameters
        ----------
        f : BIDSFile
            The BIDSFile instance to match against.

        Returns
        -------
        the matched value if a match was found, otherwise None.
        """
        ...
    def unique(self):  # -> list[Unknown]:
        """Return all unique values/levels for the current entity."""
        ...
    def count(self, files=...):  # -> int:
        """Return a count of unique values or files.

        Parameters
        ----------
        files : bool
            When True, counts all files mapped to the Entity.
            When False, counts all unique values.

        Returns
        -------
        int
            Count of unique values or files.
        """
        ...

class Tag(Base):
    """Represents an association between a File and an Entity.

    Parameters
    ----------
    file : BIDSFile
        The associated BIDSFile.
    entity : Entity
        The associated Entity.
    value : json-serializable type
        The value to store for this file/entity pair. Must be of type
        str, int, float, bool, or any json-serializable structure.
    dtype : str
        Optional type for the value field. If None, inferred from
        value. If passed, must be one of str, int, float, bool, or json.
        Any other value will be treated as json (and will fail if the
        value can't be serialized to json).
    is_metadata : bool
        Indicates whether or not the Entity is derived
        from JSON sidecars (True) or is a predefined Entity from a
        config (False).
    """

    __tablename__ = ...
    file_path = ...
    entity_name = ...
    _value = ...
    _dtype = ...
    is_metadata = ...
    file = ...
    entity = ...
    def __init__(self, file, entity, value, dtype=..., is_metadata=...) -> None: ...
    def __repr__(self): ...

class FileAssociation(Base):
    __tablename__ = ...
    src = ...
    dst = ...
    kind = ...

config_to_entity_map = ...



import itertools as it
from collections.abc import Iterable, Sequence
from typing import TypeVar

import more_itertools as itx

T = TypeVar("T")


def unpack(iterable: Iterable[T], default: Sequence[T]) -> Iterable[T]:
    """Return default if iterable has no elements.

    Allows safe unpacking of possibly empty iterables

    Parameters
    ----------
    iterable : Iterable[T]
        Iterable to unpack
    default : Sequence[T]
        Values to return if then length of iterable is 0

    Returns
    -------
    Iterable[T]
    """
    first, second = it.tee(iterable, 2)
    if itx.ilen(first):
        return second
    return default


def drop(n: int, iterable: Iterable[T]) -> Iterable[T]:
    """Return all items of an iterable except the first *n* as a list.

    >>> drop(7, range(10))
    [7, 8, 9]

    If there are fewer than *n* items in the iterable, none of them are returned

    >>> drop(8, range(5))
    []
    """
    _list = list(iterable)
    return list(it.islice(_list, n, len(_list)))



"""
This type stub file was generated by pyright.
"""

from bids_validator import BIDSValidator

from .index import BIDSLayoutIndexer
from .layout import BIDSLayout, Query
from .models import (
    BIDSDataFile,
    BIDSFile,
    BIDSImageFile,
    BIDSJSONFile,
    Config,
    Entity,
    Tag,
)
from .utils import add_config_paths, parse_file_entities

__all__ = [
    "BIDSLayout",
    "BIDSLayoutIndexer",
    "BIDSValidator",
    "add_config_paths",
    "parse_file_entities",
    "BIDSFile",
    "BIDSImageFile",
    "BIDSDataFile",
    "BIDSJSONFile",
    "Config",
    "Entity",
    "Tag",
    "Query",
]



from __future__ import annotations

import argparse
import copy
import functools as ft
import itertools as it
import re
from pathlib import Path
from typing import Any, ClassVar

import pytest
from hypothesis import given
from hypothesis import strategies as st

import tests.strategies as sb_st
from snakebids.plugins.component_edit import ComponentEdit, FilterParseError
from snakebids.types import InputsConfig, OptionalFilter
from tests.helpers import allow_function_scoped


class TestAddArguments:
    mock_args_special: ClassVar[list[str]] = ["--derivatives", "path/to/nowhere"]
    mock_basic_args: ClassVar[list[str]] = [
        "script_name",
        "path/to/input",
        "path/to/output",
        "participant",
    ]
    mock_all_args: ClassVar[list[str]] = mock_basic_args + mock_args_special

    @given(key=st.text())
    def test_works_when_key_not_present(self, key: str):
        comp_edit = ComponentEdit(components_key=key)
        p = argparse.ArgumentParser()
        comp_edit.add_cli_arguments(p, {})
        assert len(p._actions) == 1

    @given(key=st.text())
    @allow_function_scoped
    def test_config_key_can_be_set(self, key: str):
        comp_edit = ComponentEdit(components_key=key)
        p = argparse.ArgumentParser()
        comp_edit.add_cli_arguments(p, {key: {"comp": {}}})
        assert len(p._actions) == 4  # noqa: PLR2004

    @given(sb_st.inputs_configs())
    @allow_function_scoped
    def test_dynamic_inputs(self, pybids_inputs: InputsConfig):
        comp_edit = ComponentEdit()
        p = argparse.ArgumentParser()
        comp_edit.add_cli_arguments(p, {"pybids_inputs": pybids_inputs})
        magic_filters = list(
            it.chain.from_iterable(
                [[f"--filter-{key}", "entity=value"] for key in pybids_inputs]
            )
        )
        magic_wildcards = list(
            it.chain.from_iterable(
                [[f"--wildcards-{key}", "test"] for key in pybids_inputs]
            )
        )
        magic_path = list(
            it.chain.from_iterable([[f"--path-{key}", "test"] for key in pybids_inputs])
        )
        argv = magic_filters + magic_wildcards + magic_path
        args = p.parse_args(argv)
        for key in pybids_inputs:
            assert args.__dict__[f"{comp_edit.PREFIX}.path.{key}"] == "test"
            assert args.__dict__[f"{comp_edit.PREFIX}.filter.{key}"] == {
                "entity": "value"
            }
            assert args.__dict__[f"{comp_edit.PREFIX}.wildcards.{key}"] == ["test"]

    @given(
        pybids_inputs=sb_st.inputs_configs(),
        flag=st.from_regex(
            re.compile(r"(?:required)|(?:any)", re.IGNORECASE), fullmatch=True
        ),
    )
    @allow_function_scoped
    def test_required_filters(self, pybids_inputs: InputsConfig, flag: str):
        p = argparse.ArgumentParser()
        comp_edit = ComponentEdit()
        comp_edit.add_cli_arguments(p, {"pybids_inputs": pybids_inputs})
        argv = list(
            it.chain.from_iterable(
                [[f"--filter-{key}", f"entity:{flag}"] for key in pybids_inputs]
            )
        )

        args = p.parse_args(argv)
        for key in pybids_inputs:
            assert args.__dict__[f"{comp_edit.PREFIX}.filter.{key}"]["entity"] is True

    @given(
        pybids_inputs=sb_st.inputs_configs(),
        flag=st.from_regex(re.compile(r"optional", re.IGNORECASE), fullmatch=True),
    )
    @allow_function_scoped
    def test_optional_filters(self, pybids_inputs: InputsConfig, flag: str):
        p = argparse.ArgumentParser()
        comp_edit = ComponentEdit()
        comp_edit.add_cli_arguments(p, {"pybids_inputs": pybids_inputs})
        argv = list(
            it.chain.from_iterable(
                [[f"--filter-{key}", f"entity:{flag}"] for key in pybids_inputs]
            )
        )

        args = p.parse_args(argv)
        for key in pybids_inputs:
            assert (
                args.__dict__[f"{comp_edit.PREFIX}.filter.{key}"]["entity"]
                is OptionalFilter
            )

    @given(
        pybids_inputs=sb_st.inputs_configs(),
        flag=st.from_regex(
            re.compile(r"(?:match)|(?:search)", re.IGNORECASE), fullmatch=True
        ),
        value=st.text(),
    )
    @allow_function_scoped
    def test_regex_filter(self, pybids_inputs: InputsConfig, flag: str, value: str):
        p = argparse.ArgumentParser()
        comp_edit = ComponentEdit()
        comp_edit.add_cli_arguments(p, {"pybids_inputs": pybids_inputs})
        argv = list(
            it.chain.from_iterable(
                [[f"--filter-{key}", f"entity:{flag}={value}"] for key in pybids_inputs]
            )
        )

        args = p.parse_args(argv)
        for key in pybids_inputs:
            assert (
                args.__dict__[f"{comp_edit.PREFIX}.filter.{key}"]["entity"][
                    flag.lower()
                ]
                == value
            )

    @given(
        pybids_inputs=sb_st.inputs_configs(),
        flag=st.from_regex(re.compile(r"none", re.IGNORECASE), fullmatch=True),
    )
    @allow_function_scoped
    def test_none_filters(self, pybids_inputs: InputsConfig, flag: str):
        p = argparse.ArgumentParser()
        comp_edit = ComponentEdit()
        comp_edit.add_cli_arguments(p, {"pybids_inputs": pybids_inputs})
        argv = list(
            it.chain.from_iterable(
                [[f"--filter-{key}", f"entity:{flag}"] for key in pybids_inputs]
            )
        )

        args = p.parse_args(argv)
        for key in pybids_inputs:
            assert args.__dict__[f"{comp_edit.PREFIX}.filter.{key}"]["entity"] is False

    @given(
        pybids_inputs=sb_st.inputs_configs(min_size=1, max_size=1),
        flag=st.text().filter(
            lambda s: s.lower()
            not in {"none", "any", "optional", "required", "match", "search"}
            and "=" not in s
        ),
        value=st.just("") | st.text().map(lambda s: f"={s}"),
    )
    def test_filter_with_invalid_spec(
        self, pybids_inputs: InputsConfig, flag: str, value: str
    ):
        p = argparse.ArgumentParser()
        comp_edit = ComponentEdit()
        comp_edit.add_cli_arguments(p, {"pybids_inputs": pybids_inputs})
        argv = list(
            it.chain.from_iterable(
                [[f"--filter-{key}", f"entity:{flag}{value}"] for key in pybids_inputs]
            )
        )

        with pytest.raises(
            FilterParseError,
            match=re.compile(r"is not a valid filter method"),
        ):
            p.parse_args(argv)

    @given(
        pybids_inputs=sb_st.inputs_configs(min_size=1, max_size=1),
        flag=st.sampled_from(["match", "search"]),
    )
    def test_filter_with_missing_value_errors(
        self, pybids_inputs: InputsConfig, flag: str
    ):
        p = argparse.ArgumentParser()
        comp_edit = ComponentEdit()
        comp_edit.add_cli_arguments(p, {"pybids_inputs": pybids_inputs})
        argv = list(
            it.chain.from_iterable(
                [[f"--filter-{key}", f"entity:{flag}"] for key in pybids_inputs]
            )
        )

        with pytest.raises(
            FilterParseError,
            match=re.compile(rf"':{flag}' requires a value"),
        ):
            p.parse_args(argv)

    @given(
        pybids_inputs=sb_st.inputs_configs(min_size=1, max_size=1),
        flag=st.sampled_from(["none", "any", "optional", "required"]),
        value=st.text(),
    )
    def test_boolean_filter_with_value_errors(
        self, pybids_inputs: InputsConfig, flag: str, value: str
    ):
        p = argparse.ArgumentParser()
        comp_edit = ComponentEdit()
        comp_edit.add_cli_arguments(p, {"pybids_inputs": pybids_inputs})
        argv = list(
            it.chain.from_iterable(
                [[f"--filter-{key}", f"entity:{flag}={value}"] for key in pybids_inputs]
            )
        )

        with pytest.raises(
            FilterParseError,
            match=re.compile(rf"'entity:{flag}' should not be given a value"),
        ):
            p.parse_args(argv)

    @given(
        pybids_inputs=sb_st.inputs_configs(min_size=1, max_size=1),
        filt=st.text().filter(
            lambda s: "=" not in s and ":" not in s and not s.startswith("-")
        ),
    )
    def test_filters_with_only_key_errors(self, pybids_inputs: InputsConfig, filt: str):
        p = argparse.ArgumentParser()
        comp_edit = ComponentEdit()
        comp_edit.add_cli_arguments(p, {"pybids_inputs": pybids_inputs})
        argv = list(
            it.chain.from_iterable([[f"--filter-{key}", filt] for key in pybids_inputs])
        )

        with pytest.raises(
            FilterParseError,
            match=re.compile(r"Filters must be specified as"),
        ):
            p.parse_args(argv)


@st.composite
def two_configs(
    draw: st.DrawFn, filters: bool = True, wildcards: bool = True, paths: bool = True
) -> tuple[InputsConfig, InputsConfig]:
    inputs_configs = ft.partial(
        sb_st.inputs_configs, filters=filters, wildcards=wildcards, paths=paths
    )
    first = draw(inputs_configs())
    if not first:
        return first, {}
    second = draw(inputs_configs(keys=st.sampled_from(list(first.keys()))))
    return first, second


class TestUpdateNamespace:
    def make_namespace(self, config: InputsConfig):
        namespace: dict[str, Any] = {}
        for comp, conf in config.items():
            if "filters" in conf:
                namespace[f"{ComponentEdit.PREFIX}.filter.{comp}"] = conf["filters"]
            if "wildcards" in conf:
                namespace[f"{ComponentEdit.PREFIX}.wildcards.{comp}"] = conf[
                    "wildcards"
                ]
            if "custom_path" in conf:
                namespace[f"{ComponentEdit.PREFIX}.path.{comp}"] = conf["custom_path"]
        return namespace

    def test_works_when_key_not_present(self):
        comp_edit = ComponentEdit()
        config: dict[Any, Any] = {}
        comp_edit.update_cli_namespace({}, config)
        assert config == {}

    @given(configs=two_configs(wildcards=False, paths=False))
    def test_filters_are_joined_with_preference_to_namespace(
        self, configs: tuple[InputsConfig, InputsConfig]
    ):
        comp_edit = ComponentEdit()
        config = configs[0]
        namespace_orig = configs[1]
        namespace = self.make_namespace(configs[1])
        comp_edit.update_cli_namespace(namespace, {"pybids_inputs": config})
        for comp, conf in namespace_orig.items():
            for entity, filter_ in conf.get("filters", {}).items():
                assert config[comp].get("filters", {})[entity] == filter_

    @given(configs=two_configs(wildcards=False, paths=False))
    def test_original_filters_remain_intact(
        self, configs: tuple[InputsConfig, InputsConfig]
    ):
        comp_edit = ComponentEdit()
        config = configs[0]
        config_orig = copy.deepcopy(config)
        namespace_orig = configs[1]
        namespace = self.make_namespace(configs[1])
        comp_edit.update_cli_namespace(namespace, {"pybids_inputs": config})
        for comp, conf in config_orig.items():
            for entity, filter_ in conf.get("filters", {}).items():
                if entity not in namespace_orig.get(comp, {}).get("filters", {}):
                    assert config[comp].get("filters", {})[entity] == filter_

    @given(config=sb_st.inputs_configs(wildcards=False, paths=False))
    def test_optional_filter_removes_filters(self, config: InputsConfig):
        comp_edit = ComponentEdit()
        namespace_orig = {
            comp: {"filters": dict.fromkeys(conf.get("filters", {}), OptionalFilter)}
            for comp, conf in config.items()
        }
        namespace = self.make_namespace(namespace_orig)  # type: ignore
        comp_edit.update_cli_namespace(namespace, {"pybids_inputs": config})
        for conf in config.values():
            assert conf.get("filters", {}) == {}

    @given(configs=two_configs(filters=False, paths=False))
    def test_wildcards_are_appended(self, configs: tuple[InputsConfig, InputsConfig]):
        comp_edit = ComponentEdit()
        config = configs[0]
        config_orig = copy.deepcopy(config)
        namespace_orig = configs[1]
        namespace = self.make_namespace(configs[1])
        comp_edit.update_cli_namespace(namespace, {"pybids_inputs": config})
        for comp, conf in config.items():
            assert conf.get("wildcards", []) == config_orig[comp].get(
                "wildcards", []
            ) + namespace_orig.get(comp, {}).get("wildcards", [])

    @given(configs=two_configs(filters=False, wildcards=False))
    def test_path_is_overwritten_and_resolved(
        self, configs: tuple[InputsConfig, InputsConfig]
    ):
        comp_edit = ComponentEdit()
        config = configs[0]
        config_orig = copy.deepcopy(config)
        namespace_orig = configs[1]
        namespace = self.make_namespace(configs[1])
        comp_edit.update_cli_namespace(namespace, {"pybids_inputs": config})
        for comp, conf in config.items():
            if (
                ref_path := namespace_orig.get(comp, {}).get("custom_path")
            ) is not None:
                custom_path = conf["custom_path"]  # type: ignore
                assert custom_path == Path(ref_path).resolve()
            else:
                assert conf.get("custom_path") == config_orig[comp].get("custom_path")

    @given(configs=two_configs())
    def test_namespace_is_cleared(self, configs: tuple[InputsConfig, InputsConfig]):
        comp_edit = ComponentEdit()
        config = configs[0]
        namespace = self.make_namespace(configs[1])
        comp_edit.update_cli_namespace(namespace, {"pybids_inputs": config})
        assert namespace == {}



"""
This type stub file was generated by pyright.
"""

if False: ...
__all__ = ["ScalarFloat", "ExponentialFloat", "ExponentialCapsFloat"]

class ScalarFloat(float):
    def __new__(cls: Any, *args: Any, **kw: Any) -> Any: ...
    def __iadd__(self, a: Any) -> Any: ...
    def __ifloordiv__(self, a: Any) -> Any: ...
    def __imul__(self, a: Any) -> Any: ...
    def __ipow__(self, a: Any) -> Any: ...
    def __isub__(self, a: Any) -> Any: ...
    @property
    def anchor(self) -> Any: ...
    def yaml_anchor(self, any: bool = ...) -> Any: ...
    def yaml_set_anchor(self, value: Any, always_dump: bool = ...) -> None: ...
    def dump(self, out: Any = ...) -> None: ...

class ExponentialFloat(ScalarFloat):
    def __new__(cls, value: Any, width: Any = ..., underscore: Any = ...) -> Any: ...

class ExponentialCapsFloat(ScalarFloat):
    def __new__(cls, value: Any, width: Any = ..., underscore: Any = ...) -> Any: ...



import json

import jinja2
from jinja2.ext import Extension


def toml_string(item: str):
    """Encode string for inclusion in toml.

    Technically encodes as json, a (mostly) strict subset of toml, with some encoding
    fixes
    """
    return json.dumps(item, ensure_ascii=False).replace("\x7f", "\\u007f")


def toml_encode(item: str):
    """Encode string for inclusion in toml without wrapping in quotes.

    Technically encodes as json, a (mostly) strict subset of toml, with some encoding
    fixes
    """
    return toml_string(item)[1:-1]


class TomlEncodeExtension(Extension):
    """Enable the toml_string filter, which encodes provided value as toml."""

    def __init__(self, env: jinja2.Environment):
        env.filters["toml_string"] = toml_string  # type: ignore
        env.filters["toml_encode"] = toml_encode  # type: ignore



"""Recompile the bids function stub file based on latest specs."""

from __future__ import annotations

import inspect
import itertools as it
import re
from collections.abc import Iterable
from pathlib import Path
from types import ModuleType

from snakebids.paths import _config, specs
from snakebids.paths._templates import spec_func
from snakebids.paths._utils import BidsPathSpecFile, get_specs


def generate_stub(mod: ModuleType, imports: list[str], funcs: Iterable[str]):
    """Write stub file to provided module.

    Parameters
    ----------
    mod
        module for which stub file should be written
    imports
        list of imports and other statements to appear at the beginning of the file
    funcs
        list of function declarations
    """
    prelude = [
        "# This stub file is automatically generated",
        "# It can be updated using::",
        "#",
        "#      poetry run poe update_bids",
        "",
    ]
    if mod.__file__ is None:
        msg = f"{mod} has no associated file"
        raise FileNotFoundError(msg)
    Path(mod.__file__).with_suffix(".pyi").write_text(
        "\n".join(it.chain(prelude, imports, funcs))
    )


SOURCE_TEMPLATE = """
# <AUTOUPDATE>
# The code between these tags is automatically generated. Do not
# manually edit
# To update, run::
#
#       poetry run poe update-bids
#
{all_block}
{statements}
# </AUTOUPDATE>
"""

ALL_TEMPLATE = """
if not TYPE_CHECKING:
    __all__ = [ # noqa:F822
        {all}
    ]

    def __dir__():
        return __all__
"""


def update_source(
    mod: ModuleType,
    all_items: Iterable[str] | None,
    statements: list[str] | None = None,
):
    """Update python source file with __all__ declaration and provided statements.

    Parameters
    ----------
    mod
        Module to update
    all_items
        Items to include in __all__ declaration
    statements
        Statements to include after __all__ declaration block
    """
    source = inspect.getsource(mod)
    compiled = re.compile(
        r"#\s?<AUTOUPDATE>(?:.*\n)+#\s?<\/AUTOUPDATE>", flags=re.MULTILINE
    )
    if not compiled.search(source):
        err = f"Could not find an existing <AUTOUPDATE> block in {mod.__name__}"
        raise ValueError(err)
    all_block = (
        ALL_TEMPLATE.strip().format(all=",".join(f'"{item}"' for item in all_items))
        if all_items is not None
        else ""
    )
    replaced = compiled.sub(
        SOURCE_TEMPLATE.strip().format(
            all_block=all_block, statements="\n".join(statements or [])
        ),
        source,
    )
    Path(inspect.getfile(mod)).write_text(replaced)


def spec_stub(versions: Iterable[BidsPathSpecFile], latest: BidsPathSpecFile):
    """Generate spec function stub declarations.

    Parameters
    ----------
    versions
        Versions for which to generate declarations
    latest
        Version to which the `latest` stub should point
    """
    for version in versions:
        yield spec_func.format_pyi(version)

    yield spec_func.format_pyi(latest)


def get_latest(versions: Iterable[BidsPathSpecFile]) -> tuple[str, BidsPathSpecFile]:
    """Get the latest version according to semver order."""
    all_versions: dict[str, BidsPathSpecFile] = {}

    for version in versions:
        all_versions[version["version"]] = version

    version = sorted(all_versions, key=lambda v: tuple(v.split(".")))[-1]
    return version.replace(".", "_"), {
        **all_versions[version],
        "version": "latest",
    }


def main():
    """update_bids entrypoint."""
    all_specs = sorted(get_specs(), key=lambda o: o["version"])
    latest_version, latest_spec = get_latest(all_specs)
    generate_stub(
        specs,
        ["from ._utils import BidsPathSpec", "LATEST: str"],
        spec_stub(all_specs, latest_spec),
    )

    versions = [spec["version"].replace(".", "_") for spec in all_specs]
    spec_list = ",".join(f'"{v}"' for v in versions)
    valid_specs = spec_list + ', "latest"'
    update_source(_config, None, [f"VALID_SPECS: TypeAlias = Literal[{valid_specs}]"])

    update_source(
        specs,
        it.chain((version for version in versions), ("latest", "LATEST")),
        [f"_SPECS = [{spec_list}]", f'# LATEST = "{latest_version}"'],
    )


if __name__ == "__main__":
    main()



"""
This type stub file was generated by pyright.
"""

import datetime

if False: ...

class TimeStamp(datetime.datetime):
    def __init__(self, *args: Any, **kw: Any) -> None: ...
    def __new__(cls, *args: Any, **kw: Any) -> Any: ...
    def __deepcopy__(self, memo: Any) -> Any: ...
    def replace(
        self,
        year: Any = ...,
        month: Any = ...,
        day: Any = ...,
        hour: Any = ...,
        minute: Any = ...,
        second: Any = ...,
        microsecond: Any = ...,
        tzinfo: Any = ...,
        fold: Any = ...,
    ) -> Any: ...



from __future__ import annotations

import argparse
import warnings
from collections.abc import Mapping
from typing import Any

from snakebids.types import InputsConfig


def add_dynamic_args(
    parser: argparse.ArgumentParser,
    parse_args: Mapping[str, Any],
    pybids_inputs: InputsConfig,
) -> None:
    """Do nothing.

    Originally added --filter-<comp> and --wildcards-<comp> arguments to the CLI. Kept
    as a placeholder for apps that relied on it for generating documentation. This
    functionality is now native to `SnakeBidsApp`.
    """
    warnings.warn(
        "add_dynamic_args() is deprecated and no longer has any effect. Its function "
        "is now provided natively by `SnakeBidsApp`. It will be removed in an upcoming "
        "release",
        stacklevel=2,
    )



"""
This type stub file was generated by pyright.
"""

"""Functionality related to validation of BIDSLayouts and BIDS projects."""
MANDATORY_BIDS_FIELDS = ...
MANDATORY_DERIVATIVES_FIELDS = ...
EXAMPLE_BIDS_DESCRIPTION = ...
EXAMPLE_DERIVATIVES_DESCRIPTION = ...
DEFAULT_LOCATIONS_TO_IGNORE = ...

def absolute_path_deprecation_warning(): ...
def indexer_arg_deprecation_warning(): ...
def validate_root(root, validate): ...
def validate_derivative_paths(paths, layout=..., **kwargs): ...
def validate_indexing_args(ignore, force_index, root): ...



"""Tools to manage generation and interconversion of bidsapps and snakemake outputs."""

from __future__ import annotations

import json
from pathlib import Path
from typing import Literal

import more_itertools as itx

from snakebids.exceptions import RunError

Mode = Literal["workflow", "bidsapp"]


def prepare_bidsapp_output(outputdir: Path, force_output: bool = False) -> None:
    """Ensure output directory is in the correct mode and is ready for snakemake to run.

    Checks for existing output at the directory, creating it if necessary. Creates a
    .snakebids file to track output mode and other workflow information. If outputdir
    already has contents but no .snakebids file, it raises an exception unless
    force_output == True.

    Parameters
    ----------
    outputdir : Path
        Path to output
    force_output : bool
        Force output in a new output directory that already has contents. Defaults to
        False.

    Returns
    -------
    Path
        Path to new root folder (output for bidsapp, output/results for workflow)

    Raises
    ------
    RunError
        Raised when the output directory already has contents but no .snakebids file to
        identify it

    """
    # Look for .snakebids file. If the outputdir doesn't yet exist, we'll get None.
    # If it does exist but there's no .snakebids file, an error will be raised.
    try:
        _get_snakebids_file(outputdir)
    except RunError:
        if not force_output:
            raise

    outputdir.mkdir(exist_ok=True)

    write_output_mode(outputdir / ".snakebids", "bidsapp")


def write_output_mode(dotfile: Path, mode: Mode) -> None:
    """Write output mode to .snakebids.

    Parameters
    ----------
    dotfile: Path
        Path to .snakebids file to be written
    mode: Mode
        Mode to write: either "bidsapp" or "workflow"
    """
    if dotfile.exists():
        with dotfile.open("r") as f:
            data = json.load(f)
    else:
        data = {}
    data["mode"] = mode
    with (dotfile).open("w") as f:
        json.dump(data, f)


def _get_snakebids_file(outputdir: Path) -> dict[str, str] | None:
    """Ensure populated dir contains .snakebids file, retrieving it if it does.

    First checks if outputdir doesn't exist or is completely empty, returning None if
    so. If it does have data, it checks for a .snakebids file, returning its contents if
    found. If no .snakebids file is found, it raises an exception.

    Parameters
    ----------
    outputdir : Path
        Directory to check.

    Returns
    -------
    Dict or None
        None if output dir is nonexistent or empty, otherwise the contents
        of the .snakebids file

    Raises
    ------
    RunError
        Raised if outputdir contains contents but no .snakebids file

    """
    # Check if outputdir exits
    if not outputdir.exists():
        return None

    # If it does exist, is it empty?
    if itx.ilen(outputdir.iterdir()) == 0:
        return None

    # If it's not empty, is there a .snakebids file?
    if (outputdir / ".snakebids").exists():
        malformed_err = RunError(
            f"Found malformed .snakebids file in `{outputdir.resolve()}. Please "
            "remove this file and check the integrity of previous outputs."
        )
        with (outputdir / ".snakebids").open("r") as f:
            try:
                snakebids_data: dict[str, str] = json.load(f)
            except json.JSONDecodeError as err:
                raise malformed_err from err
        if "mode" not in snakebids_data:
            raise malformed_err

        return snakebids_data

    # We have an occupied directory without a .snakebids file, so we have no idea
    # what's there.
    msg = (
        f"Output dir `{outputdir.resolve()}` exists, but it doesn't look like this "
        "app has been run there before. If you're sure you got the directory correct, "
        "run the app again using `--force-output`"
    )
    raise RunError(
        msg,
    )
